import numpy as np
import pyvista as pv


def test_surface_distance():
    from ntrfc.geometry.surface import calc_dist_from_surface

    surf_one = pv.Plane()
    surf_two = pv.Plane()
    z_shift = 1.0
    surf_two.points += np.array([0, 0, z_shift])
    dist = calc_dist_from_surface(surf_one, surf_two)
    assert any(dist["distances"] == z_shift)


def test_massflow_plane():
    import numpy as np
    import pyvista as pv
    from ntrfc.geometry.plane import massflow_plane

    plane = pv.Plane()
    numcells = plane.number_of_cells
    plane["U"] = plane.cell_normals
    plane["rho"] = np.ones(numcells)

    mflow = np.sum(massflow_plane(plane))

    assert mflow == 1.0, "something is wrong"


def test_areaave_plane():
    import numpy as np
    import pyvista as pv
    from ntrfc.geometry.plane import areaave_plane

    plane = pv.Plane()
    plane["U"] = np.ones(plane.number_of_cells)

    plane_ave = areaave_plane(plane, "U")

    assert plane_ave == 1.0, "something is not right"


def test_massflowave_plane():
    import pyvista as pv
    import numpy as np
    from ntrfc.geometry.plane import massflowave_plane

    plane = pv.Plane()
    numcells = plane.number_of_cells
    plane["U"] = plane.cell_normals
    plane["rho"] = np.ones(numcells)
    plane["k"] = np.ones(numcells) * 2

    assert massflowave_plane(plane, "k") == 2.0, "something went wrong"



import numpy as np

from ntrfc.math.vectorcalc import compute_minmax_distance_in_pointcloud


def test_absVec():
    import numpy as np
    from ntrfc.math.vectorcalc import vecAbs
    a = np.array([1, 1])
    assert 2 ** .5 == vecAbs(a)
    b = np.array([7, 4, 4])
    assert 9 == vecAbs(b)


def test_vecDir():
    import numpy as np
    from ntrfc.math.vectorcalc import vecDir, vecAbs

    b = vecDir(np.array([1, 1, 1]))
    assert vecAbs(b) == 1.0


def test_randomUnitVec():
    import numpy as np
    from ntrfc.math.vectorcalc import randomUnitVec, vecAbs
    rvec = randomUnitVec()
    assert np.isclose(vecAbs(rvec), 1)


def test_RotFromTwoVecs():
    import numpy as np
    from ntrfc.math.vectorcalc import Rz
    from ntrfc.math.vectorcalc import RotFromTwoVecs
    a = np.array([1, 0, 0])
    b = np.array([0, 1, 0])

    Rab = RotFromTwoVecs(b, a)
    Rcontrol = Rz(np.pi / 2)
    assert all(np.isclose(Rab, Rcontrol).flatten())


def test_posVec():
    import numpy as np
    from ntrfc.math.vectorcalc import vecAbs, posVec

    a = np.array([-1, 0, 0])
    alength = vecAbs(a)
    b = posVec(a)
    blength = vecAbs(b)
    assert alength == blength
    assert all(np.isclose(-1 * a, b).flatten())


def test_vecProjection():
    import numpy as np
    from ntrfc.math.vectorcalc import vecProjection, vecAbs

    a = np.array([1, 0, 0])
    b = np.array([0, 1, 0])
    c = vecProjection(a, b)

    assert vecAbs(c) == 0.0

    d = np.array([1, 0, 0])
    e = np.array([1, 1, 0])
    f = vecProjection(d, e)
    assert vecAbs(f) == 1.0


def test_vecAngle():
    import numpy as np
    from ntrfc.math.vectorcalc import vecAngle
    a = np.array([1, 0, 0])
    b = np.array([0, 1, 0])
    angle = vecAngle(a, b)
    assert angle == np.pi / 2


def test_vecAbs_list():
    import numpy as np
    from ntrfc.math.vectorcalc import vecAbs_list

    a = np.array([1, 0, 0])
    b = np.array([9, 0, 0])
    c = np.stack([a, b])

    c_mags = vecAbs_list(c)
    assert c_mags[0] == 1, "inaccuracy in magnitudes!"
    assert c_mags[1] == 9, "inaccuracy in magnitudes!"


def test_unitvec_list():
    from ntrfc.math.vectorcalc import unitvec_list
    a = np.array([1, 0, 0])
    b = np.array([0, 0, 9])
    c = np.stack([a, b])
    directions = unitvec_list(c)
    assert all(np.equal(directions[0], np.array([1, 0, 0])))
    assert all(np.equal(directions[1], np.array([0, 0, 1])))


def test_unitvec():
    from ntrfc.math.vectorcalc import unitvec
    a = np.array([9, 0, 0])
    am = unitvec(a)
    assert all(np.equal(am, np.array([1, 0, 0])))
    b = np.array([0, 0, -9])
    bm = unitvec(b)
    assert all(np.equal(bm, np.array([0, 0, -1])))


def test_compute_minimal_distance_in_pointcloud():
    # Test case 3: Two points in point cloud
    pointcloud = [[1, 2, 3], [4, 5, 6]]
    assert np.isclose(compute_minmax_distance_in_pointcloud(pointcloud, minmax="min"), 5.196152, atol=1e-6)

    # Test case 4: Multiple points in point cloud
    pointcloud = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    assert np.isclose(compute_minmax_distance_in_pointcloud(pointcloud, minmax="min"), 5.196152, atol=1e-6)

    # Test case 5: Duplicate points in point cloud
    pointcloud = [[1, 2, 3], [4, 5, 6], [1, 2, 3]]
    assert np.isclose(compute_minmax_distance_in_pointcloud(pointcloud, minmax="min"), 0, atol=1e-6)

    # Test case 6: Points with negative coordinates in point cloud
    pointcloud = [[-1, -2, -3], [4, -5, 6], [-7, 8, -9]]
    assert np.isclose(compute_minmax_distance_in_pointcloud(pointcloud, minmax="min"), 10.723805294763608, atol=1e-6)

    # Test case 6: Points with negative coordinates in point cloud
    pointcloud = [[-1, -2, -3], [4, -5, 6], [-7, 8, -9]]
    assert np.isclose(compute_minmax_distance_in_pointcloud(pointcloud, minmax="max"), 22.693611435820433, atol=1e-6)



import numpy as np
import pyvista as pv

from ntrfc.math.methods import calcAnisoMatrix, calcAnisoEigs, C_barycentric, autocorr, zero_crossings, reldiff, \
    return_intersection, is_equidistant, autocorrelate, minmax_normalize
from ntrfc.timeseries.stationarity import estimate_error_jacknife


def test_ellipsoidVol():
    import numpy as np
    import pyvista as pv
    from ntrfc.math.vectorcalc import ellipsoidVol
    sigma = np.array([1, 1, 1])
    ellipsoid = pv.ParametricEllipsoid(*sigma)
    calcVol = ellipsoidVol(sigma)
    assert np.isclose(calcVol, ellipsoid.volume, rtol=1e-03, atol=1e-03)


def test_gradToRad():
    import numpy as np
    from ntrfc.math.vectorcalc import gradToRad
    angle_grad = 180
    angle_rad = gradToRad(angle_grad)
    assert np.pi == angle_rad


def test_symToMatrix():
    import numpy as np
    from ntrfc.math.vectorcalc import symToMatrix
    A = np.array([1, 1, 1, 1, 1, 1])
    R = symToMatrix(A)
    assert all(np.equal(np.ones((3, 3)), R).flatten())


def test_Rx():
    import numpy as np
    from ntrfc.math.vectorcalc import gradToRad, Rx
    angle = 90
    R = Rx(gradToRad(angle))
    test_vec = np.array([0, 0, 1])
    new_vec = np.dot(R, test_vec)
    assert all(np.isclose(new_vec, np.array([0, 1, 0])))


def test_Ry():
    import numpy as np
    from ntrfc.math.vectorcalc import gradToRad, Ry
    angle = 90
    R = Ry(gradToRad(angle))
    test_vec = np.array([1, 0, 0])
    new_vec = np.dot(R, test_vec)
    assert all(np.isclose(new_vec, np.array([0, 0, 1])))


def test_Rz():
    import numpy as np
    from ntrfc.math.vectorcalc import gradToRad, Rz
    angle = 90
    R = Rz(gradToRad(angle))
    test_vec = np.array([0, 1, 0])
    new_vec = np.dot(R, test_vec)
    assert all(np.isclose(new_vec, np.array([1, 0, 0])))


def test_lineseg():
    import pyvista as pv
    import numpy as np
    from ntrfc.math.vectorcalc import lineseg_dist
    line = pv.Line()
    testpt = np.array([0, 1, 0])
    pt_a, pt_b = line.points[0], line.points[-1]
    assert line.length == lineseg_dist(testpt, pt_a, pt_b)


def test_findNearest():
    from ntrfc.math.vectorcalc import findNearest
    import pyvista as pv
    import numpy as np
    res = 100
    line = pv.Line(resolution=res)
    point = np.array([0, 0, 0])
    near = findNearest(line.points, point)
    assert near == int(res / 2)


def test_eulersFromRPG():
    from ntrfc.math.vectorcalc import eulersFromRPG, RotFromTwoVecs, vecAngle
    import numpy as np
    a = np.array([1, 0, 0])
    b = np.array([0, 1, 0])
    cangle = vecAngle(a, b)
    R = RotFromTwoVecs(a, b)

    angle = eulersFromRPG(R)
    assert angle[0] == cangle


def test_randomOrthMat():
    from ntrfc.math.vectorcalc import randomOrthMat
    import numpy as np
    o = randomOrthMat()
    dot_o = np.dot(o, o.T)
    assert all(np.isclose(dot_o, np.identity(3)).flatten())


def test_line_intersection():
    import numpy as np
    from ntrfc.math.vectorcalc import line_intersection

    intersect = line_intersection((-1, 0), (1, 0),
                                  (0, -1), (0, 1))
    assert all(intersect == np.array([0, 0]))


def test_largedistance_indices():
    from ntrfc.math.vectorcalc import calc_largedistant_idx

    line = pv.Line(resolution=100)
    xx, yy = line.points[::, 0], line.points[::, 1]
    id1, id2 = calc_largedistant_idx(xx, yy)
    assert id1 == 0
    assert id2 == 100


def test_calcanisomatrix():
    from ntrfc.math.vectorcalc import symToMatrix
    r = np.array([1, 1, 1, 1, 1, 1])
    rm = symToMatrix(r)
    calcAnisoMatrix(rm)
    # todo implement real test


def test_calcanisoeigs():
    from ntrfc.math.vectorcalc import symToMatrix
    r = np.array([1, 1, 1, 1, 1, 1])
    rm = symToMatrix(r)
    aniso = calcAnisoMatrix(rm)
    calcAnisoEigs(aniso)
    # todo implement real test


def test_calcbarycentric():
    from ntrfc.math.vectorcalc import symToMatrix
    r = np.array([1, 1, 1, 1, 1, 1])
    rm = symToMatrix(r)
    C_barycentric(rm)
    # todo implement real test


def test_autocorr():
    time = np.linspace(0, 200, 10000)
    signal = np.sin(time)
    corr = autocorr(signal)
    zeroc = zero_crossings(corr)

    assert np.isclose(time[zeroc[0]], np.pi / 2, rtol=0.05)


def test_reldiff():
    # Test individual float values
    assert np.isclose(reldiff(20, 10), 2 / 3)
    assert np.isclose(reldiff(0, 9), 9)
    assert np.isclose(reldiff(10, 0), 10)
    assert np.isclose(reldiff(10, 10), 0)

    # Test numpy arrays
    a = np.array([[1, 2, 3], [4, 5, 6]])
    b = np.array([[1, 2, 3], [4, 5, 7]])
    assert np.allclose(reldiff(a, b), [[0, 0, 0], [0, 0, 0.15384615]])

    a = np.array([0, 0, 0])
    b = np.array([1, 2, 3])
    assert np.allclose(reldiff(a, b), [1, 2, 3])

    a = np.array([[1, 0, 3], [4, 5, 6]])
    b = np.array([[1, 2, 3], [4, 5, 7]])
    assert np.allclose(reldiff(a, b), [[0, 2, 0], [0, 0, 0.15384615]])

    # Test float and array inputs
    a = 1.0
    b = np.array([[1, 2, 3], [4, 5, 6]])
    assert np.allclose(reldiff(a, b), [[0, 2 / 3, 1], [1.2, 4 / 3, 1.42857143]])

    a = np.array([[1, 2, 3], [4, 5, 6]])
    b = 2.0
    assert np.allclose(reldiff(a, b), [[2 / 3, 0, 0.4], [2 / 3, 0.8571428571428571, 1]])


def test_return_intersection():
    hist_1 = np.array([1, 2, 3, 4])
    hist_2 = np.array([2, 3, 2, 3])
    intersection = return_intersection(hist_1, hist_2)
    assert intersection == 0.8, f"Expected 0.8, got {intersection}"

    hist_1 = np.array([1, 1, 1, 1])
    hist_2 = np.array([2, 2, 2, 2])
    intersection = return_intersection(hist_1, hist_2)
    assert intersection == 0.5, f"Expected 0.5, got {intersection}"

    hist_1 = np.array([1, 1, 1, 1])
    hist_2 = np.array([1, 1, 1, 1])
    intersection = return_intersection(hist_1, hist_2)
    assert intersection == 1.0, f"Expected 1.0, got {intersection}"


def test_is_equidistant():
    assert is_equidistant([1, 3, 5, 7, 9]) == True
    assert is_equidistant([1, 3, 5, 8, 9]) == False
    assert is_equidistant([1]) == False
    assert is_equidistant([]) == False


def test_estimate_error():
    # Generate correlated time series using sine function
    res = 10000
    timeseries = np.sin(np.linspace(0, 8 * np.pi, res)) + np.random.normal(size=res) * 0.1
    mean_error, var_error, acorr_error = estimate_error_jacknife(timeseries, block_size=res // 4)
    assert mean_error <= 0.1
    assert var_error <= 0.1
    assert acorr_error <= 0.1

    # Generate a simple Gaussian, non self-correlated distributed timeseries with a mean of 0
    timeseries = np.random.randn(res * 6)
    mean_error, var_error, acorr_error = estimate_error_jacknife(timeseries, block_size=res // 2)
    assert mean_error <= 0.2
    assert var_error <= 0.2
    assert acorr_error <= 4


def test_autocorrelate():
    # Test with a simple sinusoidal signal
    res = 48
    signal = np.sin(np.linspace(0, 4 * np.pi, res))
    expected = res / 4
    assert np.argmin(autocorrelate(signal)) == expected

    # Test with a constant signal
    signal = np.zeros(100)
    expected = 0
    assert np.argmin(autocorrelate(signal)) == expected


def test_deterministic_timescale():
    # Test with a simple sinusoidal signal
    signal = np.sin(np.linspace(0, 4 * np.pi, 100))
    expected = 1 / 2
    assert np.isclose(deterministic_timescale(signal), expected)


def deterministic_timescale(series):
    accr = autocorrelate(series)
    zcs = np.where(np.diff(np.sign(accr)))[0]
    if len(zcs) < 2:
        return False
    tsc = (zcs[1] - zcs[0]) * 2 / len(series)
    return tsc


def test_minmax_normalize():
    # Test case 1: all values are the same
    assert min(minmax_normalize(np.array([3, 3, 3, 3]))) == 3

    # Test case 3: typical input array
    input_array = np.array([1, 2, 3, 4, 5])
    expected_output = np.array([0.0, 0.25, 0.5, 0.75, 1.0])
    assert np.allclose(minmax_normalize(input_array), expected_output)

    # Test case 4: negative values in input array
    input_array = np.array([-5, -2, 0, 2, 5])
    expected_output = np.array([0., 0.3, 0.5, 0.7, 1.])
    assert np.allclose(minmax_normalize(input_array), expected_output)

    # Test case 5: positive and negative values in input array
    input_array = np.array([-3, -1, 0, 2, 5])
    expected_output = np.array([0., 0.25, 0.375, 0.625, 1.])
    assert np.allclose(minmax_normalize(input_array), expected_output)



import tempfile

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv

from ntrfc.cascade_case.casemeta.casemeta import CaseMeta
from ntrfc.cascade_case.utils.domain_utils import DomainParameters
from ntrfc.cascade_case.utils.probecontainer import ProbeContainer
from ntrfc.cascade_case.utils.sliceseries import SliceSeries
from ntrfc.filehandling.mesh import load_mesh
from ntrfc.geometry.plane import massflowave_plane
from ntrfc.math.vectorcalc import vecAngle, vecAbs
from ntrfc.turbo.bladeloading import calc_inflow_cp
from ntrfc.turbo.integrals import avdr


class GenericCascadeCase():
    """A container for data related to a cascade case, including geometry data and fluid flow data.

    This class provides functionality for reading in data from file and storing it in instance variables, as well as
    postprocessing, defining a probe proberegistry, and defining a sliceseriesregistry using inherited classes.

    Attributes:
        solver (object): An object representing the solver used to generate the data for this case.
        inlet (pv.PolyData): Geometry data for the inlet region of the case.
        outlet (pv.PolyData): Geometry data for the outlet region of the case.
        blade (pv.PolyData): Geometry data for the blade in the case.
        fluid (pv.UnstructuredGrid): Fluid flow data for the case.
        yper_low (pv.PolyData): Geometry data for the lower y-perpendicular plane of the case.
        yper_high (pv.PolyData): Geometry data for the upper y-perpendicular plane of the case.
        zper_low (pv.PolyData): Geometry data for the lower z-perpendicular plane of the case.
        zper_high (pv.PolyData): Geometry data for the upper z-perpendicular plane of the case.
        probes (ProbeContainer): A registry of probes defined for this case.
        sliceseries (SliceSeries): A registry of slices defined for this case.
    """

    def __init__(self, case_root_directory=None):
        super().__init__()
        if case_root_directory:
            self.case_meta = CaseMeta(case_root_directory)
        else:
            self.case_meta = CaseMeta(tempfile.mkdtemp())

        self.mesh_dict = {
            "inlet": pv.PolyData(),
            "outlet": pv.PolyData(),
            "blade": pv.PolyData(),
            "fluid": pv.UnstructuredGrid(),
            "yper_low": pv.PolyData(),
            "yper_high": pv.PolyData(),
            "zper_low": pv.PolyData(),
            "zper_high": pv.PolyData(),
        }

        self.sliceseries = SliceSeries()
        self.probes = ProbeContainer()
        self.domainparams = DomainParameters()
        self.active_blade_slice = pv.PolyData()

    def read_meshes(self, path, name):
        """
        Read data for any region from a file and store it in the mesh_dict.

        Args:
            path (str): Path to the file containing the geometry data.
        """

        self.mesh_dict[name] = load_mesh(path)

    def set_active_blade_slice(self, z=None):
        if not z:
            bounds = self.mesh_dict["blade"].bounds
            z = bounds[4] + (bounds[5] - bounds[4]) / 2
        self.active_blade_slice = self.mesh_dict["blade"].slice(normal="z", origin=(0, 0, z))

    def compute_domainparams_from(self, alpha):
        self.domainparams.generate_params_by_pointcloud(self.active_blade_slice, alpha)

    def compute_avdr_inout_massave(self):

        self.mesh_dict["inlet"]["u"] = self.mesh_dict["inlet"]["U"][::, 0]
        self.mesh_dict["inlet"]["v"] = self.mesh_dict["inlet"]["U"][::, 1]
        self.mesh_dict["inlet"]["w"] = self.mesh_dict["inlet"]["U"][::, 2]

        self.mesh_dict["outlet"]["u"] = self.mesh_dict["outlet"]["U"][::, 0]
        self.mesh_dict["outlet"]["v"] = self.mesh_dict["outlet"]["U"][::, 1]
        self.mesh_dict["outlet"]["w"] = self.mesh_dict["outlet"]["U"][::, 2]
        rho_1 = massflowave_plane(self.mesh_dict["inlet"], "rho")
        mag_u_1 = vecAbs(
            np.array([massflowave_plane(self.mesh_dict["inlet"], "u"), massflowave_plane(self.mesh_dict["inlet"], "v"),
                      massflowave_plane(self.mesh_dict["inlet"], "w")]))
        U_1 = np.stack(
            [massflowave_plane(self.mesh_dict["inlet"], "u"), massflowave_plane(self.mesh_dict["inlet"], "v"),
             massflowave_plane(self.mesh_dict["inlet"], "w")])
        beta_1 = vecAngle(U_1, np.array([1, 0, 0]))
        rho_2 = massflowave_plane(self.mesh_dict["outlet"], "rho")
        U_2 = np.stack(
            [massflowave_plane(self.mesh_dict["outlet"], "u"), massflowave_plane(self.mesh_dict["outlet"], "v"),
             massflowave_plane(self.mesh_dict["outlet"], "w")])
        mag_u_2 = vecAbs(np.array(
            [massflowave_plane(self.mesh_dict["outlet"], "u"), massflowave_plane(self.mesh_dict["outlet"], "v"),
             massflowave_plane(self.mesh_dict["outlet"], "w")]))
        beta_2 = vecAngle(U_2, np.array([1, 0, 0]))
        self.avdr = avdr(rho_1, mag_u_1, beta_1, rho_2, mag_u_2, beta_2)
        return self.avdr

    def blade_loading(self, figpath=tempfile.mkdtemp() + "/blade_loading.png"):
        sspoints = self.domainparams.sspoly.points
        pspoints = self.domainparams.pspoly.points

        inlet = self.mesh_dict["inlet"]
        inlet["u"] = inlet["U"][::, 0]
        inlet["v"] = inlet["U"][::, 1]
        inlet["w"] = inlet["U"][::, 2]
        p1 = massflowave_plane(inlet, valname="p", rhoname="rho", velocityname="U")
        rho = massflowave_plane(inlet, valname="rho", rhoname="rho", velocityname="U")
        u = massflowave_plane(inlet, valname="u", rhoname="rho", velocityname="U")
        v = massflowave_plane(inlet, valname="v", rhoname="rho", velocityname="U")
        w = massflowave_plane(inlet, valname="w", rhoname="rho", velocityname="U")
        U = vecAbs([u, v, w])
        pt1 = p1 + 1 / 2 * rho * U ** 2

        bladepoly = self.domainparams.profile_points

        ssmeshpointids = [bladepoly.find_closest_point(pt) for pt in sspoints]
        psmeshpointids = [bladepoly.find_closest_point(pt) for pt in pspoints]

        ssmeshpoints = bladepoly.extract_points(ssmeshpointids)
        psmeshpoints = bladepoly.extract_points(psmeshpointids)

        ind_le = self.domainparams.leading_edge_index
        ind_te = self.domainparams.trailing_edge_index
        ssmeshpoints.points -= bladepoly.points[ind_le]
        psmeshpoints.points -= bladepoly.points[ind_le]
        bladepoly.points -= bladepoly.points[ind_le]
        camber_length = vecAbs(bladepoly.points[ind_le] - bladepoly.points[ind_te])

        camber_angle = self.domainparams.stagger_angle

        ssmeshpoints.rotate_z(camber_angle, inplace=True)
        psmeshpoints.rotate_z(camber_angle, inplace=True)

        ps_xc = np.zeros(psmeshpoints.number_of_points)
        ps_cp = np.zeros(psmeshpoints.number_of_points)

        for idx, pts1 in enumerate(psmeshpoints.points):
            ps_xc[idx] = pts1[0] / camber_length
            ps_cp[idx] = calc_inflow_cp(psmeshpoints.point_data["p"][idx], pt1, p1)

        ss_xc = np.zeros(ssmeshpoints.number_of_points)
        ss_cp = np.zeros(ssmeshpoints.number_of_points)

        for idx, pts1 in enumerate(ssmeshpoints.points):
            ss_xc[idx] = pts1[0] / camber_length
            ss_cp[idx] = calc_inflow_cp(ssmeshpoints.point_data["p"][idx], pt1, p1)

        plt.figure()
        plt.title("blade loading")
        plt.scatter(ss_xc, ss_cp, label="suction side")
        plt.scatter(ps_xc, ps_cp, label="pressure side")
        plt.xlabel("$x/c_{ax}$")
        plt.ylabel("$c_{p}$")
        plt.grid()
        plt.legend()
        plt.savefig(figpath)

        return 0



import os
import sys

# Add the project directory to the Python path
sys.path.insert(0, os.path.abspath(".."))

# Project information
project = "ntrfc"
author = "Your Name"

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
language = "en"

# -- Options for HTML output ----------------------------------------------

html_theme = "alabaster"
html_theme_options = {
    "logo": "logo.png",
    "github_user": "your-github-username",
    "github_repo": "your-github-repo",
    "github_banner": True,
    "show_powered_by": False,
}

html_static_path = ["_static"]
html_css_files = ["custom.css"]

# -- Options for extensions ------------------------------------------------

# Add any additional extension-specific configuration here
# For example, to configure autodoc:
# autodoc_default_options = {
#     'members': True,
#     'undoc-members': True,
# }

# -- Options for Sphinx build ---------------------------------------------

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
html_show_sourcelink = False
html_show_sphinx = False
html_favicon = "_static/favicon.ico"

# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = "ntrfcdoc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

latex_documents = [
    (master_doc, "ntrfc.tex", "ntrfc Documentation", "Your Name", "manual"),
]

# -- Options for manual page output ---------------------------------------

man_pages = [
    (master_doc, "ntrfc", "ntrfc Documentation", [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        master_doc,
        "ntrfc",
        "ntrfc Documentation",
        author,
        "ntrfc",
        "One line description of project.",
        "Miscellaneous",
    ),
]



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 24 23:33:53 2020

@author: malte
"""

import numpy as np
import pyvista as pv
import vtk
from scipy.spatial import KDTree

from ntrfc.math.vectorcalc import vecAbs_list, unitvec


def cellDirections(cellUMean, wallNorm):
    x = unitvec(cellUMean)  # mainDirection
    y = unitvec(wallNorm)  # tangential
    z = unitvec(np.cross(x, y))  # spanwise
    return np.array([x, y, z])


def cellSpans(solutionMesh, calcFrom):
    cell_spans = []
    cellIds = np.arange(0, solutionMesh.number_of_cells)
    for cellIdx in cellIds:
        wallNormal = solutionMesh["wallNormal"][cellIdx]
        uMean = solutionMesh[calcFrom][cellIdx]

        cellDirs = cellDirections(uMean, wallNormal)
        xx = cellDirs[0]
        yy = cellDirs[1]
        zz = cellDirs[2]
        cellfaces = solutionMesh.extract_cells(cellIdx).extract_surface()
        facecenters = cellfaces.cell_centers().points
        spans = facecenters[:, np.newaxis] - facecenters
        spans_x = np.abs(np.dot(spans, xx))
        spans_y = np.abs(np.dot(spans, yy))
        spans_z = np.abs(np.dot(spans, zz))
        x_span, y_span, z_span = np.max(spans_x), np.max(spans_y), np.max(spans_z)

        cell_spans.append([x_span, y_span, z_span])
    return cell_spans


def get_wall_shear_stress_velocity(mesh, dynamic_viscosity, density_fieldname, velocity_fieldname):
    """
    Calculate the wall shear stress velocity at various points on a surface in a CFD simulation.

    Parameters:
    mesh (pv.PolyData): A mesh that represents the solution_utils of the CFD simulation.
    dynamic_viscosity (float): The dynamic viscosity of the fluid.
    density_fieldname (str): The name of the field that contains the density of the fluid.
    velocity_fieldname (str): The name of the field that contains the velocity of the fluid.

    Returns:
    wall_shear_stress_velocity (np.ndarray): An array containing the velocity at which a fluid layer adjacent to
    the surface would need to move in order to experience the same shear stress as the actual fluid layer in contact
    with the surface.
    """

    grad_velocity = mesh[f"grad_{velocity_fieldname}"].reshape(mesh.number_of_cells, 3, 3)
    wall_normals = mesh["wallNormal"]
    velocity_gradient_normal = vecAbs_list(
        [np.dot(grad_velocity[i], wall_normals[i]) for i in range(mesh.number_of_cells)])
    fluid_density = mesh[density_fieldname]
    tau_w = dynamic_viscosity / fluid_density * velocity_gradient_normal
    return np.sqrt(tau_w / fluid_density)


def constructWallMesh(surfaces):
    wall = pv.UnstructuredGrid()
    for surf in surfaces:
        wall = wall.merge(surf)
    return wall


def compute_scalar_gradient(mesh, arrayname):
    mesh = mesh.compute_derivative(arrayname)
    mesh[f"grad_{arrayname}"] = mesh["gradient"]
    return mesh


def calc_dimensionless_yplus(volmesh, surfaces, use_velfield, use_rhofield, mu_0):
    surfaceMesh = constructWallMesh(surfaces)
    surfacenormals_surface = surfaceMesh.extract_surface().compute_normals()

    volmesh = compute_scalar_gradient(volmesh, use_velfield)
    walladjacentids = volmesh.find_containing_cell(surfacenormals_surface.cell_centers().points)
    volmesh_walladjacent = volmesh.extract_cells(walladjacentids)
    volmesh_walladjacent["cellCenters"] = volmesh_walladjacent.cell_centers().points
    volmesh_walladjacent["wallNormal"] = [
        surfacenormals_surface.point_data["Normals"][surfacenormals_surface.find_closest_point(i)]
        for i in volmesh_walladjacent.points]
    distcompute = pv.PolyData(volmesh_walladjacent["cellCenters"]).compute_implicit_distance(
        surfaceMesh.extract_surface())
    volmesh_walladjacent["cellCentersWallDistance"] = distcompute["implicit_distance"]
    volmesh_walladjacent["uTaus"] = get_wall_shear_stress_velocity(volmesh_walladjacent, mu_0, use_rhofield,
                                                                   use_velfield)
    YPlus = np.abs(volmesh_walladjacent["cellCentersWallDistance"]) * volmesh_walladjacent["uTaus"] / mu_0
    volmesh_walladjacent["yPlus"] = YPlus
    return volmesh_walladjacent


def calc_dimensionless_gridspacing(volmesh, surfaces, use_velfield, use_rhofield, mu_0):
    """
    :param volmesh: pyvista-vtk object
    :param surfaces: pyvista-vtk object
    :param use_velfield: string, name of the velocity field array
    :param use_rhofield:  string, name of the density field array
    :param mu_0: float. kinematic viscosity
    :return: volmesh_walladjacent: pyvista-vtk object with the nondimensionals
    """

    print("constructing surfacemesh from wall meshes ...")
    surfaceMesh = constructWallMesh(surfaces)
    print("calculating wall-normal vectors...")
    surfacenormals_surface = surfaceMesh.extract_surface().compute_normals()

    print("finding walladjacent cells")
    merged_mesh = surfacenormals_surface + volmesh
    # Compute cell-based derivative of a vector field
    derivatives = compute_scalar_gradient(merged_mesh, "U")

    derivatives = derivatives.ctp()
    # Extract the volumetric mesh of the derivative-field
    cell_types = merged_mesh.celltypes
    volumetric_cells = np.where(cell_types == vtk.VTK_HEXAHEDRON)[0]
    face_cells = np.where(cell_types == vtk.VTK_QUAD)[0]
    # Extract the derivative-field
    volmesh = derivatives.extract_cells(volumetric_cells)
    facemesh = derivatives.extract_cells(face_cells).ctp()
    walladjacentids = volmesh.find_containing_cell(surfacenormals_surface.cell_centers().points)
    volmesh_walladjacent = volmesh.extract_cells(walladjacentids)
    volmesh_walladjacent["cellCenters"] = volmesh_walladjacent.cell_centers().points

    facemesh = facemesh.extract_surface().compute_normals()
    # Extract the cell centers of the wall-adjacent cells and the surface cells
    volmesh_walladjacent_centers = volmesh_walladjacent.cell_centers().points
    surfacemesh_surface_centers = facemesh.cell_centers().points
    facemesh = facemesh.ptc()
    # Construct a KDTree from the surface cell centers and their normals
    surface_normals = facemesh["Normals"]
    surface_gradients = facemesh["grad_U"]
    tree = KDTree(surfacemesh_surface_centers)

    # Find the indices of the closest surface cell centers to each wall-adjacent cell center
    distances, indices = tree.query(volmesh_walladjacent_centers)

    # Assign the corresponding surface normals to the wall-adjacent cells
    volmesh_walladjacent["wallNormal"] = surface_normals[indices]
    volmesh_walladjacent["wallUGradient"] = surface_gradients[indices]

    print("calculating cell spans from WallNormals...")
    spanS = cellSpans(volmesh_walladjacent, use_velfield)
    volmesh_walladjacent["xSpan"] = np.array([i[0] for i in spanS])  # calculate cell span in flow direction
    volmesh_walladjacent["ySpan"] = np.array([i[1] for i in spanS])  # calculate cell span in wall normal direction
    volmesh_walladjacent["zSpan"] = np.array([i[2] for i in spanS])  # calculate cell span in span direction

    print("calculating wall-shear and friction-velocity")
    volmesh_walladjacent = volmesh_walladjacent.ptc()
    grad_velocity = volmesh_walladjacent["wallUGradient"].reshape(volmesh_walladjacent.number_of_cells, 3, 3)
    wall_normals = volmesh_walladjacent["wallNormal"]
    velocity_gradient_normal = vecAbs_list(
        [np.dot(grad_velocity[i], wall_normals[i]) for i in range(volmesh_walladjacent.number_of_cells)])
    fluid_density = volmesh_walladjacent[use_rhofield]
    u_tau = np.sqrt(mu_0 * velocity_gradient_normal / fluid_density)
    # uTaus = volmesh_walladjacent["uTaus"]
    d_v = mu_0 / volmesh_walladjacent[use_rhofield] / u_tau
    volmesh_walladjacent["DeltaXPlus"] = volmesh_walladjacent[
                                             "xSpan"] / d_v  # calculate cell span in flow direction
    volmesh_walladjacent["DeltaYPlus"] = volmesh_walladjacent[
                                             "ySpan"] / d_v  # calculate cell span in wall normal direction
    volmesh_walladjacent["DeltaZPlus"] = volmesh_walladjacent[
                                             "zSpan"] / d_v  # calculate cell span in span direction

    return volmesh_walladjacent



import numpy as np

from ntrfc.turbo.integrals import avdr


def test_avdr():
    rho_1 = 1
    mag_u_1 = 1
    beta_1 = np.pi
    rho_2 = 1
    mag_u_2 = 1
    beta_2 = np.pi

    ans = avdr(rho_1, mag_u_1, beta_1, rho_2, mag_u_2, beta_2)
    assert ans == 1, "error"

    rho_1 = 1
    mag_u_1 = 10
    beta_1 = np.pi
    rho_2 = 1
    mag_u_2 = 20
    beta_2 = np.pi

    ans = avdr(rho_1, mag_u_1, beta_1, rho_2, mag_u_2, beta_2)
    assert ans == 2, "error"

    rho_1 = 1
    mag_u_1 = 10
    beta_1 = np.pi / 4
    rho_2 = 1
    mag_u_2 = 4.082485
    beta_2 = np.pi / 3

    ans = avdr(rho_1, mag_u_1, beta_1, rho_2, mag_u_2, beta_2)
    assert ans == 0.5000002566283092, "error"



import numpy as np
import pyvista as pv

from ntrfc.geometry.alphashape import calc_concavehull, auto_concaveHull
from ntrfc.geometry.line import polyline_from_points, refine_spline
from ntrfc.math.vectorcalc import vecAngle
from ntrfc.turbo.profile_tele_extraction import extract_vk_hk


def mid_length(ind_1, ind_2, sorted_poly):
    """
    calc length of a midline. currently only used in the iterative computation of LE and TE index of a profile. probably
    this method is not necessary, as it is only two lines
    :param ind_1: index LE
    :param ind_2: index TE
    :param sorted_poly: pv.PolyData sorted
    :return: length
    """

    ps_poly, ss_poly = extractSidePolys(ind_1, ind_2, sorted_poly)
    mids_poly = midline_from_sides(ps_poly, ss_poly)
    return mids_poly.length


def midline_from_sides(ps_poly, ss_poly):
    x_ps, y_ps = ps_poly.points[::, 0], ps_poly.points[::, 1]
    x_ss, y_ss = ss_poly.points[::, 0], ss_poly.points[::, 1]
    z = ps_poly.points[0][2]
    midsres = 100
    if x_ps[0] > x_ps[-1]:
        ax, ay = refine_spline(x_ps[::-1], y_ps[::-1], midsres)
    else:
        ax, ay = refine_spline(x_ps, y_ps, midsres)
    if x_ss[0] > x_ss[-1]:
        bx, by = refine_spline(x_ss[::-1], y_ss[::-1], midsres)
    else:
        bx, by = refine_spline(x_ss, y_ss, midsres)
    xmids, ymids = ((ax + bx) / 2, (ay + by) / 2)

    midsPoly = polyline_from_points(np.stack((xmids, ymids, z * np.ones(len(ymids)))).T)
    return midsPoly


def extractSidePolys(ind_1, ind_2, sortedPoly, verbose=False):
    # xs, ys = list(sortedPoly.points[::, 0]), list(sortedPoly.points[::, 1])
    indices = np.arange(0, sortedPoly.number_of_points)

    if ind_2 > ind_1:
        side_one_idx = indices[ind_1:ind_2 + 1]
        side_two_idx = np.concatenate((indices[:ind_1 + 1][::-1], indices[ind_2:][::-1]))
    elif ind_1 > ind_2:
        side_one_idx = indices[ind_2:ind_1 + 1]
        side_two_idx = np.concatenate((indices[:ind_2 + 1][::-1], indices[ind_1:][::-1]))

    side_one = extract_points_fromsortedpoly(side_one_idx, sortedPoly)
    side_two = extract_points_fromsortedpoly(side_two_idx, sortedPoly)

    side_one_spline = polyline_from_points(side_one.points)
    side_two_spline = polyline_from_points(side_two.points)

    if side_one_spline.length > side_two_spline.length:
        psPoly = side_two
        ssPoly = side_one
    else:
        psPoly = side_one
        ssPoly = side_two

    # if verbose:
    #     p = pv.Plotter()
    #     p.add_mesh(ssPoly, color="g", label="ssPoly")
    #     p.add_mesh(psPoly, color="b", label="psPoly")
    #     p.add_mesh(sortedPoly.points[ind_1], color="w", label="ind_1")
    #     p.add_mesh(sortedPoly.points[ind_2], color="k", label="ind_2")
    #     p.add_legend()
    #     p.show()

    return ssPoly, psPoly


def extract_points_fromsortedpoly(sorted_indices, sorted_poly):
    """
    Extract a subset of points and associated data from a sorted polygon.

    Parameters
    ----------
    sorted_indices : numpy.ndarray
        1D array of shape `(n_points,)` containing the indices of the points
        in `sorted_poly.points` that should be extracted and used to create a
        new polygon. The indices should be sorted in the order that they
        should appear in the new polygon.
    sorted_poly : pyvista.PolyData
        Polygon represented as a `PolyData` object from the `pyvista`
        library. The `sorted_poly.points` attribute should be a NumPy array
        of shape `(n_points, 3)` containing the 3D coordinates of the polygon
        vertices, and the `sorted_poly.point_data` attribute should be a
        dictionary of NumPy arrays containing additional per-vertex data.

    Returns
    -------
    pyvista.PolyData
        A new `PolyData` object representing the subset of the original polygon
        specified by `sorted_indices`, with all per-vertex data preserved.

    Notes
    -----
    This function assumes that the input polygon is sorted in a specific way.
    Specifically, the `sorted_indices` parameter should contain the indices of
    the points in the order that they should appear in the new polygon. If the
    input polygon is not sorted in this way, the resulting polygon may not be
    valid.
    """

    side_two = pv.PolyData(sorted_poly.points[sorted_indices])
    for arr in sorted_poly.array_names:
        side_two[arr] = sorted_poly.point_data[arr][sorted_indices]
    return side_two


def extract_geo_paras(polyblade, alpha=None, verbose=False):
    """
    This function is extracting profile-data as stagger-angle, midline, psPoly, ssPoly and more from a set of points
    Be careful, you need a suitable alpha-parameter in order to get the right geometry
    The calculation of the leading-edge and trailing-edge index needs time and its not 100% reliable (yet)
    Keep in mind, to check the results!
    :param polyblade: pyvista polymesh of the blade
    :param alpha: nondimensional alpha-coefficient (calcConcaveHull)
    :param verbose: bool for plots
    :return: points, psPoly, ssPoly, ind_vk_, ind_vk, midsPoly, beta_leading, beta_trailing, alpha
    """
    points = polyblade.points

    if alpha:
        xs, ys = calc_concavehull(points[:, 0], points[:, 1], alpha)
    else:
        xs, ys, alpha = auto_concaveHull(points[:, 0], points[:, 1])

    index_sort = [np.where(points[:, 0] == xs[i])[0][0] for i in range(len(xs)) if
                  len(np.where(points[:, 0] == xs[i])) == 1 and np.where(points[:, 0] == xs[i])[0][0] == np.where(
                      points[:, 1] == ys[i])[0][0]]

    sortedPoly = pv.PolyData(polyblade.points[index_sort])  # polyblade.extract_cells(index_sort)
    for arr in polyblade.array_names:
        if sortedPoly.number_of_points == len(polyblade[arr]):
            sortedPoly[arr] = polyblade.point_data[arr][index_sort]

    ind_vk, ind_hk = extract_vk_hk(sortedPoly)
    psPoly, ssPoly = extractSidePolys(ind_hk, ind_vk, sortedPoly)
    midsPoly = midline_from_sides(psPoly, ssPoly)

    # compute angles from 2d-midline
    xmids, ymids = midsPoly.points[::, 0], midsPoly.points[::, 1]
    vk_tangent = np.stack((xmids[0] - xmids[1], ymids[0] - ymids[1], 0)).T
    hk_tangent = np.stack((xmids[-2] - xmids[-1], ymids[-2] - ymids[-1], 0)).T
    chord = psPoly.points[0] - psPoly.points[-1]
    beta_leading = vecAngle(vk_tangent, -np.array([1, 0, 0])) / np.pi * 180
    beta_trailing = vecAngle(hk_tangent, -np.array([1, 0, 0])) / np.pi * 180
    camber_angle = vecAngle(chord, -np.array([1, 0, 0])) / np.pi * 180

    # if verbose:
    #     p = pv.Plotter()
    #     p.add_mesh(points, color="orange", label="points")
    #     p.add_mesh(psPoly, color="green", label="psPoly")
    #     p.add_mesh(ssPoly, color="black", label="ssPoly")
    #     p.add_mesh(midsPoly, color="black", label="midsPoly")
    #     p.add_mesh(pv.Line((0, 0, 0), (midsPoly.length, 0, 0)))
    #     p.add_mesh(sortedPoly.points[ind_hk], color="red", point_size=5)
    #     p.add_mesh(sortedPoly.points[ind_vk], color="orange", point_size=5)
    #     p.add_legend()
    #     p.show()

    return sortedPoly, psPoly, ssPoly, ind_vk, ind_hk, midsPoly, beta_leading, beta_trailing, camber_angle, alpha


def calcMidPassageStreamLine(x_mcl, y_mcl, beta1, beta2, x_inlet, x_outlet, t, verbose=False):
    """
    Calculate the midpoint stream line curve through a passage.

    Parameters:
    -----------
    x_mcl : array_like
        The x-coordinates of the mid-chord line.
    y_mcl : array_like
        The y-coordinates of the mid-chord line.
    beta1 : float
        The angle in degrees representing the inflow angle at the inlet.
    beta2 : float
        The angle in degrees representing the outflow angle at the outlet.
    x_inlet : float
        The x-coordinate of the inlet position.
    x_outlet : float
        The x-coordinate of the outlet position.
    t : float
        The pitch of the midpoint stream line.
    verbose : bool, optional
        If True, a plot of the midpoint stream line is displayed.

    Returns:
    --------
    x_mpsl_ref : array_like
        The refined x-coordinates of the midpoint stream line.
    y_mpsl_ref : array_like
        The refined y-coordinates of the midpoint stream line.
    """

    delta_x_vk = x_mcl[0] - x_inlet
    delta_y_vk = np.tan(np.deg2rad(beta1)) * delta_x_vk

    p_inlet_x = x_mcl[0] - delta_x_vk
    p_inlet_y = y_mcl[0] - delta_y_vk

    delta_x_hk = x_outlet - x_mcl[-1]
    delta_y_hk = delta_x_hk * np.tan(np.deg2rad(-beta2))

    p_outlet_x = x_mcl[-1] + delta_x_hk
    p_outlet_y = y_mcl[-1] + delta_y_hk

    x_mpsl = [p_inlet_x] + list(x_mcl) + [p_outlet_x]
    y_mpsl = [p_inlet_y] + list(y_mcl) + [p_outlet_y]

    for i in range(len(x_mpsl)):
        y_mpsl[i] = y_mpsl[i] + 0.5 * t

    x_mpsl_ref, y_mpsl_ref = refine_spline(x_mpsl, y_mpsl, 1000)

    return x_mpsl_ref, y_mpsl_ref



import numpy as np
import pyvista as pv


def compute_expansion_factors(grid: pv.UnstructuredGrid) -> np.ndarray:
    # Compute cell volumes
    cell_volumes = grid.compute_cell_sizes()["Volume"]

    # Number of cells in the grid
    n_cells = grid.n_cells

    # Initialize array to store expansion factors
    expansion_factors = np.zeros(n_cells)

    # Iterate over cells in the grid
    for i in range(n_cells):
        # Find the indices of the neighbors of the current cell
        neighbor_indices = get_neighbor_cell_ids(grid, i)

        # Find the volume of the largest neighboring cell
        max_neighbor_volume = max([cell_volumes[j] for j in neighbor_indices])

        # Compute expansion factor for the current cell
        expansion_factors[i] = max_neighbor_volume / cell_volumes[i]

    return expansion_factors


def get_neighbor_cell_ids(grid, cell_idx):
    """Helper to get neighbor cell IDs."""
    cell = grid.GetCell(cell_idx)
    pids = pv.vtk_id_list_to_array(cell.GetPointIds())
    neighbors = set(grid.extract_points(pids)["vtkOriginalCellIds"])
    neighbors.discard(cell_idx)
    return np.array(list(neighbors))



import os

import numpy as np
import pyvista as pv

from ntrfc.turbo.domaingen_cascade import cascade_2d_domain

ON_CI = 'CI' in os.environ


def test_cascade_3d_domain():
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.turbo.domaingen_cascade import cascade_3d_domain
    from ntrfc.turbo.domaingen_cascade import cascade_2d_domain

    if ON_CI:
        pv.start_xvfb()

    naca_code = "6509"
    angle = 30  # deg
    res = 420
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs[:-1], ys[:-1], np.zeros(len(xs) - 1)]).T)
    sorted_poly.rotate_z(angle, inplace=True)

    sortedPoly, psPoly, ssPoly, per_y_upper, per_y_lower, inletPoly, outletPoly = cascade_2d_domain(sorted_poly, -1, 2,
                                                                                                    1,
                                                                                                    "m", 0.1, 1,
                                                                                                    path=False)

    cascade_3d_domain(sortedPoly, psPoly, ssPoly, per_y_upper, per_y_lower, inletPoly, outletPoly, zspan=0.2, avdr=1,
                      path=False)


def test_cascade_2d_domain():
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca

    if ON_CI:
        pv.start_xvfb()

    naca_code = "6509"
    angle = 30  # deg
    res = 420
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs[:-1], ys[:-1], np.zeros(len(xs) - 1)]).T)
    sorted_poly.rotate_z(angle, inplace=True)
    X, Y = sorted_poly.points[::, 0], sorted_poly.points[::, 1]
    points = np.stack((X[:], Y[:], np.zeros(len(X)))).T
    pointspoly = pv.PolyData(points)
    sortedPoly, psPoly, ssPoly, per_y_upper, per_y_lower, inletPoly, outletPoly = cascade_2d_domain(pointspoly, -1, 2,
                                                                                                    1,
                                                                                                    "m", 0.1, 1, False)



import numpy as np
import pandas as pd
from scipy import signal

from ntrfc.math.methods import minmax_normalize


def optimal_bin_width(sample1, sample2):
    """
    Compute the optimal bin width using cross-validation estimator for mean integrated squared error.

    Parameters:
    -----------
    data : numpy array
        One-dimensional array of data points.

    Returns:
    --------
    h : float
        Optimal bin width.
    """

    data = np.concatenate([sample1, sample2])
    if np.all(data == data[0]):
        return [data[0] - 0.5, data[0] + 0.5]
    n = len(data)
    h = 2 * np.std(data) * n ** (-1 / 3)  # initial bin width using Scott's rule

    # perform cross-validation to estimate mean integrated squared error
    J_min = np.inf
    for i in range(12):
        bins = np.arange(np.min(data), np.max(data) + h, h)
        counts, _ = np.histogram(data, bins=bins)
        N_k = counts[np.nonzero(counts)]
        J = 2 / ((n - 1) * h) - (n + 1) / (n ** 2 * (n - 1) * h) * np.sum(N_k ** 2)
        if J < J_min:
            J_min = J
            h_opt = h
        h *= 0.8  # decrease bin width for better accuracy
    bins = np.arange(min(np.min(sample1), np.min(sample2)), max(np.max(sample1), np.max(sample2)), h_opt)
    return bins


def smd_probability_compare(sample1, sample2, verbose=False):
    """Compare the probability distribution of two signals using the Freedman-Diaconis rule
    to determine the number of bins.

    Args:
        sample1 (numpy.ndarray): First signal.
        sample2 (numpy.ndarray): Second signal.

    Returns:
        float: Mean squared error between the probability distribution densities
            of the two signals. A value of 0 indicates that the probability distributions
            are not alike, while a value of 1 indicates that they are equal.
    """
    # Compute the number of bins using the Freedman-Diaconis rule.
    bins = optimal_bin_width(sample1, sample2)

    # Compute the histogram and probability density of the first signal.
    hist1, _ = np.histogram(sample1, bins=bins, density=True)
    pdf1 = hist1 / np.sum(hist1)

    # Compute the histogram and probability density of the second signal.
    hist2, _ = np.histogram(sample2, bins=bins, density=True)
    pdf2 = hist2 / np.sum(hist2)

    # Compute the mean squared error between the probability densities.
    mse = np.sum((pdf1 - pdf2) ** 2)
    # Convert the mse to a similarity score between 0 and 1.
    similarity = 1 - mse
    # if verbose:
    #     # Plot the histogram
    #     plt.hist(sample1, bins=bins)
    #     plt.hist(sample2, bins=bins)
    #     plt.show()
    return similarity


def optimal_window_size(time_series, min_interval=0.05, max_interval=0.25, verbose=False):
    """
    Determines the optimal window size for a given time series.

    Parameters:
    ----------
    time_series : array-like
        The time series to analyze.
    verbose : bool, optional
        If True, a plot of the correlation coefficient and KS test results for each window size will be displayed.

    Returns:
    -------
    int or bool
        The optimal window size for the time series. If no suitable window size is found, False is returned.

    Notes:
    -----
    The function normalizes the input time series using the minmax_normalize() function.
    The window size is chosen based on a cumulative score that takes into account the correlation coefficient and
     KS test p-value.
    The function returns False if no suitable window size is found, meaning the input time series does not exhibit the
     necessary periodicity.
    """

    # Normalize the time series
    normalized_series = minmax_normalize(time_series)

    # Get the length of the time series and define a range of allowed window sizes
    series_length = len(normalized_series)
    allowed_window_sizes = np.array(range(int(series_length * min_interval), int(series_length * max_interval) + 1))

    # Iterate through the allowed window sizes and perform the Kolmogorov-Smirnov test and correlation coefficient calculation
    mean_scores = []
    var_scores = []
    for window_size in allowed_window_sizes:
        check_window = normalized_series[-window_size * 2:]
        check_window_df = pd.DataFrame(check_window)
        rolling_check = check_window_df.rolling(window_size)
        mean_scores.append(np.std(rolling_check.mean()).values[0])
        var_scores.append(np.std(rolling_check.var()).values[0])
        # Compute the correlation coefficient
    cumulated_scores = minmax_normalize(mean_scores) + minmax_normalize(np.array(var_scores))

    optimal_window_size_index = np.argmin(cumulated_scores)
    opt_window_size = allowed_window_sizes[optimal_window_size_index]

    #
    opt_window = time_series[-opt_window_size * 2:]

    assert len(opt_window) == opt_window_size * 2
    probability_similiarity = smd_probability_compare(opt_window[:opt_window_size], opt_window[opt_window_size:])

    if probability_similiarity < 0.96:
        return False, False, False

    # Compute the period of the time series
    freqs, psd = signal.welch(opt_window.T, fs=1, nperseg=opt_window_size // 2)
    maxpsdid = np.argmax(psd)

    if maxpsdid != 0:
        tperiod = freqs[np.argmax(psd)] ** -1
        nperiods = (opt_window_size + (-opt_window_size % tperiod)) // tperiod
    else:
        tperiod = np.inf
        nperiods = 0

    # If verbose mode is enabled, display a plot of the correlation coefficient and KS test results for each window size
    # if verbose:
    #     plt.plot(cumulated_scores)
    #     plt.axvline(optimal_window_size_index)
    #     plt.legend()
    #     plt.show()

    return opt_window, opt_window_size, nperiods


def estimate_stationarity(timeseries, verbose=False):
    sigma_threshold = 3
    normalized_series = minmax_normalize(timeseries)  # minmax_normalize(timeseries)
    datalength = len(normalized_series)
    opt_window, opt_window_size, nperiods = optimal_window_size(normalized_series)
    if not opt_window_size:
        return False

    reference_window = opt_window

    reference_mean = np.mean(reference_window)
    reference_variance = np.var(reference_window)
    opt_rolling_window = np.lib.stride_tricks.sliding_window_view(opt_window, opt_window_size)

    rolling_means = np.mean(opt_rolling_window, axis=1)
    rolling_vars = np.var(opt_rolling_window, axis=1)

    assert len(rolling_means) == len(opt_rolling_window)
    assert len(rolling_vars) == len(opt_rolling_window)

    mean_uncertainty = np.std(rolling_means)
    var_uncertainty = np.std(rolling_vars)

    checkseries = normalized_series[:-opt_window_size * 2]

    checkseries_reversed = pd.DataFrame(checkseries[::-1])
    rolling_win_reversed = checkseries_reversed.rolling(window=opt_window_size)

    rolling_means_reversed = rolling_win_reversed.mean().values
    rolling_vars_reversed = rolling_win_reversed.var().values

    outer_mean_uncertainty = 0.0015
    outer_var_uncertainty = outer_mean_uncertainty * 0.25  # variance can only be 25% of minmax normed series

    rolling_means_errors_reversed = np.abs(rolling_means_reversed - reference_mean)
    rolling_vars_errors_reversed = np.abs(rolling_vars_reversed - reference_variance)

    mean_limits = sigma_threshold * (mean_uncertainty) + outer_mean_uncertainty
    var_limits = sigma_threshold * (var_uncertainty) + outer_var_uncertainty

    rolling_means_errors_inliers_reversed = rolling_means_errors_reversed[opt_window_size:] <= mean_limits
    rolling_vars_errors_inliers_reversed = rolling_vars_errors_reversed[opt_window_size:] <= var_limits

    def last_coherent_interval(arr, threshold=1):
        reversed_arr = arr[::-1]
        false_indices = np.where(reversed_arr == False)[0]
        last_false_index = false_indices[-1] if len(false_indices) > 0 else 0
        coherent_arr = [False] * last_false_index + [True] * (len(reversed_arr) - last_false_index)
        success_rate = np.array([np.sum(coherent_arr[i:]) / len(coherent_arr[i:]) for i in range(len(coherent_arr))])
        answer_index = np.where(success_rate >= threshold)[0][0]
        return len(arr) - answer_index

    mean_index = datalength - last_coherent_interval(rolling_means_errors_inliers_reversed) - 3 * opt_window_size

    variance_index = datalength - last_coherent_interval(rolling_vars_errors_inliers_reversed) - 3 * opt_window_size

    stationary_start_index = max(mean_index, variance_index)

    # if verbose:
    #     fig, axs = plt.subplots(3, 1, figsize=(24, 20))
    #     axs[0].plot(normalized_series, label="normalized series", color="blue")
    #     axs[0].vlines(x=stationary_start_index, ymin=0, ymax=np.nanmax(normalized_series), label="stationary_start",
    #                   color="k")
    #
    #     axs[1].plot(np.nan_to_num(rolling_means_errors_reversed[::-1], 0), label="mean error", color="red")
    #     axs[1].hlines(y=(mean_limits), xmin=0, xmax=len(normalized_series), label="mean_limits", color="k")
    #     axs[1].vlines(x=stationary_start_index, ymin=0, ymax=max(mean_limits, np.nanmax(rolling_means_errors_reversed)),
    #                   label="stationary_start", color="k")
    #     axs[1].vlines(x=mean_index, ymin=0, ymax=max(mean_limits, np.nanmax(rolling_means_errors_reversed)),
    #                   label="mean_index", color="green")
    #     axs[1].legend()
    #
    #     axs[2].plot(np.nan_to_num(rolling_vars_errors_reversed[::-1], 0), label="variance error", color="red")
    #     axs[2].hlines(y=(var_limits), xmin=0, xmax=len(normalized_series), label="var_limits", color="k")
    #     axs[2].vlines(x=stationary_start_index, ymin=0, ymax=max(var_limits, np.nanmax(rolling_vars_errors_reversed)),
    #                   label="stationary_start", color="k")
    #     axs[2].vlines(x=variance_index, ymin=0, ymax=max(var_limits, np.nanmax(rolling_vars_errors_reversed)),
    #                   label="variance_index", color="k")
    #     axs[2].legend()
    #
    #     plt.show()

    return stationary_start_index


def estimate_error_jacknife(timeseries, block_size=20, n_samples=4000):
    """
    Estimates the errors of the mean, variance, and autocorrelation of a given time series using jackknife resampling method.

    Parameters
    ----------
    timeseries : array-like
        The input time series.
    block_size : int, optional
        The block size used in the jackknife resampling method (default is 20).
    n_samples : int, optional
        The number of jackknife samples to generate (default is 4000).

    Returns
    -------
    tuple
        A tuple of three floats representing the error estimates of the mean, variance, and autocorrelation, respectively.

    Notes
    -----
    The jackknife resampling method is used to estimate the errors of the mean, variance, and autocorrelation of the
     input time series.
    The function generates `n_samples` jackknife samples by randomly selecting blocks from the time series calculates
     the mean, variance, and autocorrelation of each jackknife sample.
    It also generates `n_samples` noise samples with the same block size as the original time series and calculates the
     mean, variance, and autocorrelation of each noise sample.
    The standard deviation of the jackknife estimates for mean, variance, and autocorrelation are calculated, and each
     is multiplied by a factor of 16 to obtain the final error estimates.

    Choosing an appropriate block size is crucial for obtaining reliable and accurate estimates of the errors of the
     mean, variance, and autocorrelation of a given time series using the jackknife resampling method.
    """

    # Original time series

    x = timeseries
    #
    n_blocks = len(timeseries) // block_size

    # Initialize arrays to store jackknife estimates
    mean_jk = np.zeros(n_samples)
    var_jk = np.zeros(n_samples)
    acorr_jk = np.zeros(n_samples)

    for i in range(n_samples):
        # Generate a random index array of block indices
        idx = np.random.randint(0, n_blocks, size=n_blocks)
        # Select blocks according to the random indices
        start = idx * block_size
        end = start + block_size
        x_jk = np.concatenate([x[start[i]:end[i]] for i in range(len(start))])

        # Calculate the mean, variance, and autocorrelation of the jackknife sample
        mean_jk[i] = np.mean(x_jk)
        var_jk[i] = np.var(x_jk)

        freqs, psd = signal.periodogram(x_jk, return_onesided=False)
        acorr_jk[i] = freqs[np.argmax(psd)]

    mean_jk_error = np.std(mean_jk)
    var_jk_error = np.std(var_jk)
    accr_jk_error = np.std(acorr_jk)

    return mean_jk_error, var_jk_error, accr_jk_error



import os

import pyvista as pv


def read_vtk(path_to_mesh):
    """Read a vtk/vtp/vtu file and return the corresponding mesh object.

    Parameters
    ----------
    path_to_mesh : str
        The file path to the vtk/vtp/vtu file to be read.

    Returns
    -------
    mesh : pyvista.UnstructuredGrid or pyvista.StructuredGrid
        The mesh object constructed from the input file.
    """
    return pv.read(path_to_mesh)


def read_vtm(path_to_mesh):
    """Read a vtm file and return the corresponding mesh object.

    Parameters
    ----------
    path_to_mesh : str
        The file path to the vtm file to be read.

    Returns
    -------
    mesh : pyvista.UnstructuredGrid or pyvista.StructuredGrid
        The combined mesh object constructed from all the blocks in the vtm file.
    """
    multi_block_mesh = pv.read(path_to_mesh)
    return multi_block_mesh.combine()


def read_cgns(path_to_mesh):
    """Read a cgns file and return the corresponding mesh object.

    Parameters
    ----------
    path_to_mesh : str
        The file path to the cgns file to be read.

    Returns
    -------
    mesh : pyvista.UnstructuredGrid or pyvista.StructuredGrid
        The combined mesh object constructed from all the unstructured or structured grids in the cgns file.
    """
    cgns_reader = pv.get_reader(path_to_mesh)
    cgns = cgns_reader.read()
    nbase = cgns.n_blocks
    basenames = [cgns.get_block_name(i) for i in range(nbase)]
    bases = [cgns[b] for b in basenames]
    mesh = pv.UnstructuredGrid()
    for base in bases:
        ndomains = base.n_blocks
        domainnames = [base.get_block_name(i) for i in range(ndomains)]
        domains = [base[i] for i in domainnames]
        if all(isinstance(grid, pv.UnstructuredGrid) for grid in domains):
            for domain in domains:
                if type(domain) == pv.UnstructuredGrid:
                    mesh = mesh.merge(domain)
                if type(domain) == pv.StructuredGrid:
                    mesh = mesh.merge(domain)
        else:
            # not all elements are UnstructuredGrid objects
            for domain in domains:
                nblocks = domain.n_blocks
                blocks = [domain[b] for b in range(nblocks)]
                for block in blocks:
                    if type(block) == pv.UnstructuredGrid:
                        mesh = mesh.merge(block)
                    if type(block) == pv.StructuredGrid:
                        mesh = mesh.merge(block)
    return mesh


def load_mesh(path_to_mesh):
    """Load a mesh file and return the corresponding mesh object.

    Parameters
    ----------
    path_to_mesh : str
        The file path to the mesh file to be loaded. The file extension should be one of ".vtk", ".vtp", ".vtu", ".vtm", ".cgns", ".msh".

    Returns
    -------
    mesh : pyvista.UnstructuredGrid or pyvista.StructuredGrid
        The mesh object constructed from the input file.
    """
    assert os.path.isfile(path_to_mesh), f"{path_to_mesh} is not a valid file"
    extension = os.path.splitext(path_to_mesh)[1]
    if extension == ".vtk" or extension == ".vtp" or extension == ".vtu":
        mesh = read_vtk(path_to_mesh)
    elif extension == ".vtm":
        mesh = read_vtm(path_to_mesh)
    elif extension == ".cgns":
        mesh = read_cgns(path_to_mesh)
    elif extension == ".msh":
        mesh = pv.read(path_to_mesh)
    return mesh



# -*- coding: utf-8 -*-
"""
Created on Sun Oct  4 19:01:50 2020

@author: malte
"""

import math as m
import sys

import numpy as np
from scipy.spatial import distance
from scipy.spatial.distance import squareform, pdist, cdist
from scipy.stats import special_ortho_group


def closest_node_index(node, nodes):
    closest_index = distance.cdist([node], nodes).argmin()
    return closest_index


def distant_node_index(node, nodes):
    closest_index = distance.cdist([node], nodes).argmax()
    return closest_index


def calc_largedistant_idx(x_koords, y_koords):
    """
    tested method to find indices of coordinates of a 2d-pointcloud with the biggest distance
    :param x_koords: array of xcords
    :param y_koords: array of ycords
    :return: index_1,index_2 (int)
    """
    A = np.dstack((x_koords, y_koords))[0]
    D = squareform(pdist(A))
    #    N = np.max(D)
    I = np.argmax(D)
    I_row, I_col = np.unravel_index(I, D.shape)

    index_1 = I_row
    index_2 = I_col

    return index_1, index_2


def symToMatrix(symTensor):
    """
    tested translates symmetric tensor notation to complete matrix
    :param symTensor:
    :return:
    """
    # xx,xy,xz,yy,yz,zz
    Matrix = np.array([[symTensor[0], symTensor[1], symTensor[2]],
                       [symTensor[1], symTensor[3], symTensor[4]],
                       [symTensor[2], symTensor[4], symTensor[5]]])
    return Matrix


def gradToRad(angle):
    """
    tested method to translate from grad to rad
    :param angle:
    :return:
    """
    return (angle / 180) * np.pi


def Rx(xAngle):
    """
    using radiant
    :param xAngle: angle in rad
    :return: rotation matrix
    """
    return np.array([[1, 0, 0],
                     [0, np.cos(xAngle), np.sin(xAngle)],
                     [0, -np.sin(xAngle), np.cos(xAngle)]])


def Ry(yAngle):
    """
    using radiant
    :param yAngle: angle in rad
    :return: rotation matrix
    """
    return np.array([[np.cos(yAngle), 0, np.sin(yAngle)],
                     [0, 1, 0],
                     [np.sin(yAngle), 0, np.cos(yAngle)]])


def Rz(zAngle):
    """
    using radiant
    :param zAngle: angle in rad
    :return: rotation matrix
    """
    return np.array([[np.cos(zAngle), np.sin(zAngle), 0],
                     [-np.sin(zAngle), np.cos(zAngle), 0],
                     [0, 0, 1]])


def RotFromTwoVecs(vec1, vec2):
    """ Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """

    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return rotation_matrix


def vecAbs(vec):
    """
    method to calculate the absolute value of a vector
    :param vec:
    :return:
    """
    return np.linalg.norm(vec)


def vecAbs_list(vecs):
    """
    method to calculate the absolute value of a vector
    :param np.array vec with shape (n,3):
    :return: array of magnitudes in shape (n)
    """
    return np.linalg.norm(vecs, axis=1)


def vecDir(vec):
    """
    tested method to compute the direction of a vector
    :param vec:
    :return: unit vec
    """
    return vec / vecAbs(vec)


def posVec(vec):
    return (vec ** 2) ** .5


def findNearest(array, point):
    nodes = np.asarray(array)
    deltas = nodes - point
    dist_2 = np.einsum('ij,ij->i', deltas, deltas)
    return np.argmin(dist_2)


def eulersFromRPG(R):
    tol = sys.float_info.epsilon * 10

    if abs(R.item(0, 0)) < tol and abs(R.item(1, 0)) < tol:
        eul1 = 0
        eul2 = m.atan2(-R.item(2, 0), R.item(0, 0))
        eul3 = m.atan2(-R.item(1, 2), R.item(1, 1))
    else:
        eul1 = m.atan2(R.item(1, 0), R.item(0, 0))
        sp = m.sin(eul1)
        cp = m.cos(eul1)
        eul2 = m.atan2(-R.item(2, 0), cp * R.item(0, 0) + sp * R.item(1, 0))
        eul3 = m.atan2(sp * R.item(0, 2) - cp * R.item(1, 2), cp * R.item(1, 1) - sp * R.item(0, 1))

    """
    print("z - phi =", eul1)
    print("y - theta  =", eul2)
    print("x - psi =", eul3)

    print("checkRPGRepo")
    print(R)

    print(np.dot(np.dot(Rz(eul1),Ry(eul2)),Rx(eul3)))
    """
    return eul1, eul2, eul3


def randomUnitVec():
    """
    tested method to generate a random unit vector
    :return:
    """
    phi = np.random.uniform(0, np.pi * 2)
    costheta = np.random.uniform(-1, 1)

    theta = np.arccos(costheta)
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    return np.array([x, y, z])


def randomOrthMat():
    num_dim = 3
    x = special_ortho_group.rvs(num_dim)
    return x


def ellipsoidVol(sig):
    """
    tested method to compute the ellipsoid volume by the sigma-parameters of a parametric ellipsoid
    :param sig:
    :return:
    """
    return 4 / 3 * np.pi * sig[0] * sig[1] * sig[2]


def vecProjection(direction, vector):
    """
    Calculate the projection of a vector onto a direction vector.

    Parameters:
    direction (list or numpy array): The direction vector onto which the projection will be calculated.
    vector (list or numpy array): The vector to be projected onto the direction vector.

    Returns:
    projection (numpy array): A vector representing the projection of the input vector onto the direction vector.
    """
    unitDir = vecDir(direction)
    return np.dot(vector, unitDir) * unitDir


def vecAngle(vec1, vec2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            angle_between((1, 0, 0), (0, 1, 0))
                1.5707963267948966
            angle_between((1, 0, 0), (1, 0, 0))
                0.0
            angle_between((1, 0, 0), (-1, 0, 0))
                3.141592653589793
    """
    absVec1 = vecAbs(vec1)
    absVec2 = vecAbs(vec2)
    return np.arccos(np.dot(vec1, vec2) / (absVec1 * absVec2))


def lineseg_dist(p, a, b):
    """
    :param p: point
    :param a: line point a
    :param b: line point b
    :return: distance
    """
    # normalized tangent vector
    d = np.divide(b - a, np.linalg.norm(b - a))

    # signed parallel distance components
    s = np.dot(a - p, d)
    t = np.dot(p - b, d)

    # clamped parallel distance
    h = np.maximum.reduce([s, t, 0])

    # perpendicular distance component
    c = np.cross(p - a, d)

    return np.hypot(h, np.linalg.norm(c))


def line_intersection(point_a1, point_a2,
                      point_b1, point_b2):
    def det_2d(a, b):
        return a[0] * b[1] - a[1] * b[0]

    xdiff = (point_a1[0] - point_a2[0], point_b1[0] - point_b2[0])
    ydiff = (point_a1[1] - point_a2[1], point_b1[1] - point_b2[1])

    div = det_2d(xdiff, ydiff)
    if div == 0:
        return None

    d = (det_2d(point_a1, point_a2), det_2d(point_b1, point_b2))
    x = det_2d(d, xdiff) / div
    y = det_2d(d, ydiff) / div
    return x, y


def unitvec(vec):
    return vec / vecAbs(vec)


def unitvec_list(vecs):
    return vecs / vecAbs_list(vecs)[:, None]


def compute_minmax_distance_in_pointcloud(pointcloud, minmax="min"):
    # Convert the pointcloud to a NumPy array
    pointcloud = np.array(pointcloud)

    # Compute the pairwise distances between points using cdist
    distances = cdist(pointcloud, pointcloud)

    if minmax == "min":
        # Exclude the diagonal elements since they represent the distance between each point and itself
        np.fill_diagonal(distances, np.inf)
        # Find the minimum distance
        distance = np.min(distances)
    elif minmax == "max":
        # Exclude the diagonal elements since they represent the distance between each point and itself
        np.fill_diagonal(distances, -np.inf)
        # Find the maximum distance
        distance = np.max(distances)

    return distance



import numpy as np
import pyvista as pv


def test_loadmesh_vtk(tmpdir):
    """
    tests if a vtk mesh can be read and Density is translated to rho
    """
    from ntrfc.filehandling.mesh import load_mesh

    test_file = tmpdir / "tmp.vtk"
    mesh = pv.Box()
    mesh["Density"] = np.ones(mesh.number_of_cells)
    mesh.save(test_file)
    mesh_load = load_mesh(test_file)
    assert "Density" in mesh_load.array_names


def test_loadmesh_cgns(tmpdir):
    from ntrfc.filehandling.mesh import load_mesh
    from pyvista.examples.downloads import download_cgns_multi
    import shutil
    download = download_cgns_multi(load=False)

    test_file = tmpdir / "tmp.cgns"
    shutil.move(download, test_file)
    mesh_load = load_mesh(test_file)
    assert "Density" in mesh_load.array_names


def test_read_vtk(tmpdir):
    from ntrfc.filehandling.mesh import read_vtk
    test_file = tmpdir / "tmp.vtk"
    # offset array.  Identifies the start of each cell in the cells array
    offset = np.array([0, 9])

    # Contains information on the points composing each cell.
    # Each cell begins with the number of points in the cell and then the points
    # composing the cell
    cells = np.array([8, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 12, 13, 14, 15])

    # cell type array. Contains the cell type of each cell
    cell_type = np.array([12, 12])

    # in this example, each cell uses separate points
    cell1 = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [1, 1, 1],
            [0, 1, 1],
        ]
    )

    cell2 = np.array(
        [
            [0, 0, 2],
            [1, 0, 2],
            [1, 1, 2],
            [0, 1, 2],
            [0, 0, 3],
            [1, 0, 3],
            [1, 1, 3],
            [0, 1, 3],
        ]
    )

    # points of the cell array
    points = np.vstack((cell1, cell2)).astype(float)

    # create the unstructured grid directly from the numpy arrays
    # The offset is optional and will be either calculated if not given (VTK version < 9),
    # or is not necessary anymore (VTK version >= 9)
    if pv.vtk_version_info < (9,):
        grid = pv.UnstructuredGrid(offset, cells, cell_type, points)
    else:
        grid = pv.UnstructuredGrid(cells, cell_type, points)

    # save the mesh to a vtk file
    grid.save(test_file)
    mesh = read_vtk(test_file)
    assert isinstance(mesh, (pv.UnstructuredGrid, pv.StructuredGrid))


#
# def test_read_vtm():
#     mesh = read_vtm(VTM_FILE)
#     assert isinstance(mesh, (pv.UnstructuredGrid, pv.StructuredGrid))

def test_read_cgns():
    from pyvista import examples
    from ntrfc.filehandling.mesh import read_cgns
    CGNS_FILE = examples.download_cgns_structured(load=False)
    mesh = read_cgns(CGNS_FILE)
    assert isinstance(mesh, (pv.UnstructuredGrid, pv.StructuredGrid))



def test_calc_concavehull():
    """
    in these simple geometries, each point must be found by calcConcaveHull
    """
    from ntrfc.geometry.alphashape import calc_concavehull
    import numpy as np
    import pyvista as pv

    square = pv.Plane()
    boxedges = square.extract_feature_edges()

    boxedges.rotate_z(np.random.randint(0, 360), inplace=True)
    boxpoints = boxedges.points

    np.random.shuffle(boxpoints)

    xs_raw = boxpoints[:, 0]
    ys_raw = boxpoints[:, 1]

    xs, ys = calc_concavehull(xs_raw, ys_raw, 1)

    assert len(xs) == len(xs_raw)
    assert any([yi in ys_raw for yi in ys])

    polygon = pv.Polygon()
    polygon.rotate_z(np.random.randint(0, 360), inplace=True)
    polyedges = polygon.extract_feature_edges()
    polypoints = polyedges.points
    np.random.shuffle(polypoints)
    xs_raw = polypoints[:, 0]
    ys_raw = polypoints[:, 1]

    xs, ys = calc_concavehull(xs_raw, ys_raw, 10)

    assert len(xs) == len(xs_raw)
    assert any([yi in ys_raw for yi in ys])


def test_calc_optimize_alphashape():
    """
    in these simple geometries, each point must be found by calcConcaveHull
    """
    from ntrfc.geometry.alphashape import auto_concaveHull
    import numpy as np
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca

    for i in range(4):
        digit_string = "6509"

        res = 240
        X, Y = naca(digit_string, res, half_cosine_spacing=True)
        points = np.stack((X[:], Y[:], np.zeros(res * 2))).T * np.random.randn() * 1000

        xs, ys, alpha = auto_concaveHull(points[::, 0], points[::, 1])

        assert len(xs) == len(X)



import numpy as np
import pyvista as pv

from ntrfc.turbo.pointcloud_methods import extract_points_fromsortedpoly, calcMidPassageStreamLine


def test_extractSidePolys():
    from ntrfc.turbo.pointcloud_methods import extractSidePolys
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    import numpy as np
    import pyvista as pv

    digit_string = "6509"

    res = 240
    X, Y = naca(digit_string, res, half_cosine_spacing=True)
    ind_hk = 0
    ind_vk = res
    points = np.stack((X[:], Y[:], np.zeros(res * 2))).T

    poly = pv.PolyData(points)
    poly["A"] = np.ones(poly.number_of_points)
    ssPoly, psPoly = extractSidePolys(ind_hk, ind_vk, poly)
    # the assertion is consistent with all tests but it is confusing
    # we generate profiles with a naca-generator. probably there is a minor bug in the algorithm
    # ssPoly needs to have one point more then psPoly
    assert ssPoly.number_of_points == psPoly.number_of_points, "number of sidepoints are not equal "


def test_extract_points_fromsortedpoly():
    # Create a sorted polygon with some per-vertex data

    sorted_poly = pv.Circle()

    # Extract a subset of the points and associated data
    sorted_indices = np.arange(sorted_poly.number_of_points // 2, sorted_poly.number_of_points)
    side_two = extract_points_fromsortedpoly(sorted_indices, sorted_poly)

    assert side_two.number_of_points == sorted_poly.number_of_points // 2


def test_calcMidPassageStreamLine():
    # Define input values
    x_mcl = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    y_mcl = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    beta1 = 10.0
    beta2 = 5.0
    x_inlet = 0.0
    x_outlet = 6.0
    t = 0.1

    # Expected output
    x_mpsl_ref_expected = np.array(
        [0., 0.00756705, 0.0151341, 0.02270115, 0.0302682, 0.03783525, 0.0454023, 0.05296935, 0.0605364, 0.06810345,
         0.07567051, 0.08323756, 0.09080461, 0.09837166, 0.10593871, 0.11350576, 0.12107281, 0.12863986, 0.13620691,
         0.14377396, 0.15134101, 0.15890806, 0.16647511])
    y_mpsl_ref_expected = np.array(
        [0.87367302, 0.87500729, 0.87634157, 0.87767584, 0.87901012, 0.8803444, 0.88167867, 0.88301295, 0.88434722,
         0.8856815, 0.88701577, 0.88835005, 0.88968432, 0.8910186, 0.89235287, 0.89368715, 0.89502142, 0.8963557,
         0.89768997, 0.89902425, 0.90035852, 0.9016928, 0.90302707, 0])

    # Calculate actual output
    x_mpsl_ref, y_mpsl_ref = calcMidPassageStreamLine(x_mcl, y_mcl, beta1, beta2, x_inlet, x_outlet, t)

    # Test output
    assert len(x_mpsl_ref) == 1000
    assert len(y_mpsl_ref) == 1000



import numpy as np


def test_p_t_is_from_mach():
    from ntrfc.fluid.isentropic import p_t_is_from_mach
    # Test with some sample inputs
    kappa = 1.4
    ma = 2
    p = 100000.0

    # Calculate expected result
    expected = 782444.9066867264

    # Calculate actual result
    actual = p_t_is_from_mach(kappa, ma, p)

    # Check if actual result matches expected result
    assert np.isclose(expected, actual)


def test_p_is_from_mach():
    from ntrfc.fluid.isentropic import p_is_from_mach
    # Test using standard values for air at sea level
    kappa = 1.4
    mach_number = 0.8
    total_pressure = 101325.0  # Pa
    expected_static_pressure = 66471.39048022314  # Pa
    calculated_static_pressure = p_is_from_mach(kappa, mach_number, total_pressure)
    assert np.isclose(calculated_static_pressure, expected_static_pressure)


def test_temp_t_is():
    from ntrfc.fluid.isentropic import temp_t_is
    assert np.isclose(temp_t_is(1.4, 0, 300), 300)
    assert np.isclose(temp_t_is(1.4, 1.0, 300), 360.0)
    assert np.isclose(temp_t_is(1.4, 1.5, 300), 435)


def test_temp_is():
    from ntrfc.fluid.isentropic import temp_is
    assert np.isclose(temp_is(1.4, 0, 340.952), 340.952)
    assert np.isclose(temp_is(1.4, 1.0, 438.298), 365.24833333333333)
    assert np.isclose(temp_is(1.4, 1.5, 578.947), 399.27379310344827)


def test_mach_is_x():
    from ntrfc.fluid.isentropic import mach_is_x
    kappa = 1.4
    p = 100000.0
    pt = 189292.91587378542
    expected_mach = 1
    mach = mach_is_x(kappa, p, pt)
    assert np.isclose(mach, expected_mach)


def test_isentropic_mach_number():
    from ntrfc.fluid.isentropic import isentropic_mach_number
    # Standard conditions
    isentropic_pressure = 101325  # Pa
    kappa = 1.4
    static_pressure = 101325  # Pa
    mach_number = 0.5
    gas_constant = 287.058  # J/(kg*K)
    static_temperature = 288.15  # K

    expected_output = 0.0014693046301270448

    # Call function and check output
    assert np.isclose(isentropic_mach_number(isentropic_pressure, kappa, static_pressure, mach_number, gas_constant,
                                             static_temperature), expected_output)



import numpy as np
import pyvista as pv
import vtk


def compute_cell_aspect_ratios(grid: pv.UnstructuredGrid) -> np.ndarray:
    """Compute the aspect ratio of each cell in an unstructured grid.

    The aspect ratio of a cell is defined as the ratio of the longest edge length
    to the shortest edge length of the cell.

    Parameters:
        grid (pv.UnstructuredGrid): The unstructured grid.

    Returns:
        np.ndarray: An array of aspect ratios, one for each cell in the grid.
    """
    cellids = range(0, grid.number_of_cells)
    # Compute the edge lengths for each cell
    aspect_ratios = []
    vtk.vtkLogger.SetStderrVerbosity(vtk.vtkLogger.VERBOSITY_OFF)
    for cellid in cellids:
        # Get the indices of the points that make up the cell
        cell = grid.extract_cells(cellid)

        edges = cell.extract_all_edges().compute_cell_sizes(length=True, area=False, volume=False)

        aspect_ratios.append(max(edges["Length"]) / min(edges["Length"]))

    return np.array(aspect_ratios)



import numpy as np
import pandas as pd

from ntrfc.math.vectorcalc import vecAngle

"""
MIT License

Copyright (c) 2020 David Poves

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://github.com/DavidPoves/11-Parameters-Turbine-Blade-Generator
"""


class TurbineBladeGen(object):
    def __init__(self, radius, axial_chord, tangential_chord, unguided_turning, inlet_blade, inlet_half_wedge, le_r,
                 outlet_blade, te_r, n_blades, throat, n_points=50):
        """
        Class implementing the parametrization method for a turbine blade based on the paper:
                        "L.J. Pritchard: An eleven parameter axial turbine airfoil geometry model"
        This method parametrizes the geometry of the airfoil based on 11 elemental parameters, which are the ones
        defined below. ANGLES SHOULD ALWAYS BE INTRODUCED IN RADIANS.
        :param radius: Radius of the cylinder upon which the airfoil is defined.
        :param axial_chord: Axial chord.
        :param tangential_chord: Tangential chord.
        :param unguided_turning: Unguided turning angle. MUST BE INTRODUCED IN RADIANS.
        :param inlet_blade: Inlet local blade angle. MUST BE INTRODUCED IN RADIANS.
        :param inlet_half_wedge: Inlet half wedge angle. MUST BE INTRODUCED IN RADIANS.
        :param le_r: Radius of the leading edge radius.
        :param outlet_blade: Outlet local blade angle. MUST BE INTRODUCED IN RADIANS.
        :param te_r: Radius of the trailing edge.
        :param n_blades: Number of blades.
        :param throat: Geometric throat.
        :param n_points: Number of points used to define the blade. Optional, default is 50.
        """
        self.R = radius
        self.cx = axial_chord
        self.ct = tangential_chord
        self.unguided = unguided_turning
        self.beta_i = inlet_blade
        self.epsilon_i = inlet_half_wedge
        self.le_r = le_r
        self.beta_o = outlet_blade
        self.te_r = te_r
        self.N = n_blades
        self.throat = throat
        self.n_points = 5 * round(n_points / 5)  # To make number of points divisible by 5.

        # Initialize parameters.
        self.x1, self.y1 = None, None
        self.x2, self.y2 = None, None
        self.x3, self.y3 = None, None
        self.x4, self.y4 = None, None
        self.x5, self.y5 = None, None
        self.x5, self.y5 = None, None
        self.x6, self.y6 = None, None
        self.x7, self.y7 = None, None
        self.x8, self.y8 = None, None
        self.x9, self.y9 = None, None
        self.x0, self.y0 = None, None
        self.R0 = 0
        self.dxp, self.dxs = None, None
        self.aS, self.bS, self.cS, self.dS = None, None, None, None
        self.aP, self.bP, self.cP, self.dP = None, None, None, None
        self.x_suction, self.y_suction = None, None
        self.x_pressure, self.y_pressure = None, None
        self.x_te_close, self.y_te_close = None, None

        # Initialize the dependent parameters.
        self._compute_dependent_parameters()

        # Get a first guess of the wedge out angle.
        self._get_first_guess_wedgeout()

        # Create a Pandas dataframe for the dependent data.
        self.DependentTable = pd.DataFrame()

    def calculate_blade_geometry(self, straight_te=True):
        """
        Compute the geometry of the blade given the 11 required geometric parameters. In order to solve for the
        discontinuity at the throat point (second key point of the airfoil) an iterative process over the wedge out
        has been implemented with an absolute tolerance of 1e-5, as indicated on the paper. An error will raise if no
        convergence is achieved. Moreover, if the user decides to implement a straight trailing edge cut (for CFD
        applications), a straight cut on the latter will be performed.
        The final geometry will finally be plotted to let the user visualize the created airfoil. Finally, the user is
        asked to choose of a .txt file with the coordinates of each of the curves should be saved. In that case, a GUI
        will be prompted to let the user select where to save the generated data.
        :param straight_te: Boolean indicating if a straight cut on the TE should be carried out.
        :return:
        """
        convergence = False  # Set the convergence parameter on the wedge out angle to solve the throat discontinuity.

        # Perform the iteration to find convergence on the wedge out angle and solve the discontinuity.
        while not convergence:
            # Compute the key points.
            self.x1, self.y1 = self._compute_first_point()
            self.x2, self.y2 = self._compute_second_point()
            self.x3, self.y3 = self._compute_third_point()
            self.x4, self.y4 = self._compute_fourth_point()
            self.x5, self.y5 = self._compute_fifth_point()

            # Compute the circular arc center and radius.
            x0 = ((self.y1 - self.y2) * np.tan(self.beta_1) * np.tan(self.beta_2) + self.x1 * np.tan(self.beta_2) -
                  self.x2 * np.tan(self.beta_1)) / ((np.tan(self.beta_2)) - np.tan(self.beta_1))
            y0 = -(x0 - self.x1) / np.tan(self.beta_1) + self.y1
            R0 = np.sqrt((self.x1 - x0) ** 2 + (self.y1 - y0) ** 2)
            yy2 = y0 + np.sqrt(R0 ** 2 - (self.x2 - x0) ** 2)

            if abs(self.y2 - yy2) < 1e-5:
                convergence = True
            elif abs(self.y2 - yy2) > 1e-5:
                self.epsilon_o *= (self.y2 / yy2) ** 4
                if self.epsilon_o < 0:  # Raise error since the angle tends to negative values.
                    raise TimeoutError(
                        'Error on removing throat discontinuity. Try reducing the exit blade angle or the throat.')

        # Assign the obtained values and compute auxiliary data points.
        self.x0, self.y0, self.R0 = x0, y0, R0
        self.x6, self.y6 = self.cx, 0.
        self.x7, self.y7 = self.cx - self.te_r, 0.
        self.x8, self.y8 = 0., self.ct
        self.x9, self.y9 = self.le_r, self.ct

        # Introduce the half wedge out angle.
        self.dependent_data['Half Wedge out Angle [degrees]'] = np.degrees(self.epsilon_o)

        # Define the first coordinate for both pressure and suction surfaces.
        x_suction, y_suction = np.array([]), np.array([])
        x_pressure, y_pressure = np.array([]), np.array([])
        x_suction, y_suction = np.append(x_suction, self.x8), np.append(y_suction, self.y8)
        x_pressure, y_pressure = np.append(x_pressure, self.x8), np.append(y_pressure, self.y8)

        # Define steps on the x coordinates for the pressure and suction sides respectively.
        self.dxp = (self.x4 - self.x8) / (
            (1 / 5) * self.n_points - 1)  # Parameter defined for the x coord step on pressure.
        self.dxs = (self.x3 - self.x8) / (
            (1 / 5) * self.n_points - 1)  # Parameter defined for the x coordinate step on suction.

        # Define the polynomial coefficients for pressure and suction sides.
        self.aS, self.bS, self.cS, self.dS = self._compute_coefficients_third_polynomial(self.x2, self.y2, self.beta_2,
                                                                                         self.x3, self.y3, self.beta_3)
        self.aP, self.bP, self.cP, self.dP = self._compute_coefficients_third_polynomial(self.x4, self.y4, self.beta_4,
                                                                                         self.x5, self.y5, self.beta_5)

        # Define the circle between 3 and 4.
        x_suction, y_suction, x_pressure, y_pressure, = self._define_pressure_suction_1(x_suction, y_suction,
                                                                                        x_pressure, y_pressure)
        # x_pressure, y_pressure = x_pressure[::-1], y_pressure[::-1]

        # Define the third order polynomials for pressure and suction surfaces.
        x_suction, y_suction, x_pressure, y_pressure, = self._define_pressure_suction_2(x_suction, y_suction,
                                                                                        x_pressure, y_pressure)

        # Define the circular arc and the remaining third order polynomial for each of the sides.
        x_suction, y_suction, x_pressure, y_pressure, = self._define_pressure_suction_3(x_suction, y_suction,
                                                                                        x_pressure, y_pressure)

        if straight_te:  # If specified, straight cut on TE.
            self.x_te_close = np.array([x_pressure[-1], x_suction[-1]])
            self.y_te_close = np.array([y_pressure[-1], y_suction[-1]])
            # Plot the data.
            # plt.plot(x_suction, y_suction, label='Suction side')
            # plt.plot(x_pressure, y_pressure, label='Pressure side')
            # plt.plot(self.x_te_close, self.y_te_close, label='Closing TE')
        else:
            x_suction, y_suction, x_pressure, y_pressure, = self._define_pressure_suction_4(x_suction, y_suction,
                                                                                            x_pressure, y_pressure)
            # # Plot the data.
            # plt.plot(x_suction, y_suction, label='Suction side')
            # plt.plot(x_pressure, y_pressure, label='Pressure side')

        # plt.xlabel('x')
        # plt.ylabel('y')
        # plt.legend()

        # Assign the data to the class.
        self.x_suction, self.y_suction = x_suction, y_suction
        self.x_pressure, self.y_pressure = x_pressure, y_pressure

        # Create the dataframe with the dependent data.
        self.DependentTable = pd.DataFrame(self.dependent_data, columns=list(self.dependent_data.keys()))
        pd.set_option('colheader_justify', 'center')  # Justify the column headers to the center.

        # # Create an input dialog to save data.
        # save_bool = input('Should data be saved (x and y coordinates; and dependent data)? [y/n]: ')
        # if save_bool == 'y':
        #     self._save_data(straight_te)

    # def _save_data(self, straight_te):
    #     np.savetxt('suction_side.txt', np.column_stack((self.x_suction, self.y_suction)))
    #     np.savetxt('pressure_side.txt', np.column_stack((self.x_pressure, self.y_pressure)))
    #     self.DependentTable.to_csv(os.path.join(os.getcwd(), 'dependent_data.csv'),
    #                                header=list(self.dependent_data.keys()), sep='\t', mode='a')
    #     if straight_te:
    #         np.savetxt('trailing_edge.txt', np.column_stack((self.x_te_close, self.y_te_close)))

    def _compute_dependent_parameters(self):
        """
        Define the parameters dependent on the geometry of the blade and the cascade, as reflected on the paper. All
        dependent angles are returned in DEGREES.
        :return:
        """
        self.pitch = (2 * np.pi * self.R) / self.N
        self.stagger = np.degrees(np.arctan(self.ct / self.cx))
        self.chord = np.sqrt(self.cx ** 2 + self.ct ** 2)
        self.Zweifel = (4 * np.pi * self.R) / (self.cx * self.N) * np.sin(self.beta_i - self.beta_o) * np.cos(
            self.beta_o) / np.cos(
            self.beta_i)
        self.solidity = self.chord / self.pitch
        self.camber_angle = np.degrees(self.beta_i - self.beta_o)
        self.Lift_coeff = (2 * self.pitch) / self.chord * 0.5 * (np.cos(self.beta_i) + np.cos(self.beta_o)) * (
            np.tan(self.beta_i) - np.tan(self.beta_o))

        # Create a dictionary with the data.
        self.dependent_data = {'Pitch': [self.pitch], 'Stagger Angle [degrees]': [self.stagger], 'Chord': [self.chord],
                               'Zweifel Coefficient': [self.Zweifel], 'Solidity': [self.solidity],
                               'Camber Angle [degrees]': [self.camber_angle], 'Lift Coefficient': [self.Lift_coeff]}

    def _get_first_guess_wedgeout(self):
        """
        Get a first guess of the wedge out angle IN RADIANS.
        :return:
        """
        self.epsilon_o = 0.5 * self.unguided  # Compute first guess on the wedge-out angle.

    def _compute_first_point(self):
        """
        Compute the first elementary point of the airfoil.
        :return: x and y coordinates of the airfoil.
        """
        beta_1 = self.beta_o - self.epsilon_o
        x1 = self.cx - self.te_r * (1 + np.sin(beta_1))
        y1 = self.te_r * np.cos(beta_1)

        # Save the value of beta for slope computations.
        self.beta_1 = beta_1

        return x1, y1

    def _compute_second_point(self):
        """
        Compute the second elementary point of the airfoil.
        :return: x and y coordinates of the second elementary point.
        """
        beta_2 = self.beta_o - self.epsilon_o + self.unguided
        x2 = self.cx - self.te_r + (self.throat + self.te_r) * np.sin(beta_2)
        y2 = (2 * np.pi * self.R) / self.N - (self.throat + self.te_r) * np.cos(beta_2)

        # Save the value of beta for slope computations.
        self.beta_2 = beta_2

        return x2, y2

    def _compute_third_point(self):
        """
        Compute the third elementary point of the airfoil.
        :return: x and y coordinates of the third elementary point.
        """
        beta_3 = self.beta_i + self.epsilon_i
        x3 = self.le_r * (1 - np.sin(beta_3))
        y3 = self.ct + self.le_r * np.cos(beta_3)

        # Save the value of beta for slope computations.
        self.beta_3 = beta_3

        return x3, y3

    def _compute_fourth_point(self):
        """
        Compute the fourth elementary point of the airfoil.
        :return: x and y coordinates of the fourth elementary point.
        """
        beta_4 = self.beta_i - self.epsilon_i
        x4 = self.le_r * (1 + np.sin(beta_4))
        y4 = self.ct - self.le_r * np.cos(beta_4)

        # Save the value of beta for slope computations.
        self.beta_4 = beta_4

        return x4, y4

    def _compute_fifth_point(self):
        """
        Compute the fifth elementary point of the airfoil.
        :return: x and y coordinates of the fifth elementary point.
        """
        beta_5 = self.beta_o + self.epsilon_o
        x5 = self.cx - self.te_r * (1 - np.sin(beta_5))
        y5 = -self.te_r * np.cos(beta_5)

        # Save the value of beta for slope computations.
        self.beta_5 = beta_5

        return x5, y5

    @staticmethod
    def _compute_coefficients_third_polynomial(x1, y1, beta_1, x2, y2, beta_2):
        """
        Compute the coefficients for a third order polynomial of the form y = a + bx + cx^2 + dx^3 given two points and
        the local blade angles at each of them.
        :param x1: x-coordinate of the first point.
        :param y1: y-coordinate of the first point.
        :param beta_1: Local blade angle at the first point. MUST BE IN RADIANS.
        :param x2: x-coordinate of the second point.
        :param y2: y-coordinate of the second point.
        :param beta_2: Local blade angle at the second point. MUST BE IN RADIANS.
        :return: Coefficients of the third order polynomial given as a tuple (a, b, c, d).
        """
        d = (np.tan(beta_1) + np.tan(beta_2)) / (x1 - x2) ** 2 - 2 * (y1 - y2) / (x1 - x2) ** 3
        c = (y1 - y2) / (x1 - x2) ** 2 - (np.tan(beta_2)) / (x1 - x2) - d * (x1 + 2 * x2)
        b = np.tan(beta_2) - 2 * c * x2 - 3 * d * x2 ** 2
        a = y2 - b * x2 - c * x2 ** 2 - d * x2 ** 3

        coefficients = (a, b, c, d)

        return coefficients

    def _define_pressure_suction_1(self, x_suction, y_suction, x_pressure, y_pressure):
        """
        Define the leading edge arc joining elementary points 3 and 4. This function will separate coords between
        pressure and suction sides.
        :param x_suction: Array containing the x-coordinates of the suction side.
        :param y_suction: Array containing the y-coordinates of the suction side.
        :param x_pressure: Array containing the x-coordinates of the pressure side.
        :param y_pressure: Array containing the y-coordinates of the pressure side.
        :return: Arrays containing the coordinates (x and y separately) for each of the sides (suction first,
        then pressure side).
        """
        for _ in np.arange(1, (1 / 5) * self.n_points + 1):
            x_pressure = np.append(x_pressure, x_pressure[-1] + self.dxp)
            y_pressure = np.append(y_pressure, self.y9 - np.sqrt(self.le_r ** 2 - (x_pressure[-1] - self.x9) ** 2))
            x_suction = np.append(x_suction, x_suction[-1] + self.dxs)
            y_suction = np.append(y_suction, self.y9 + np.sqrt(self.le_r ** 2 - (x_suction[-1] - self.x9) ** 2))
        self.dxp = (self.x5 - self.x4) / ((3 / 5) * self.n_points)
        self.dxs = (self.x2 - self.x3) / ((2 / 5) * self.n_points)

        return x_suction, y_suction, x_pressure, y_pressure

    def _define_pressure_suction_2(self, x_suction, y_suction, x_pressure, y_pressure):
        """
        Define the third order polynomials on the pressure and suction sides. This function will separate coords between
        pressure and suction sides.
        :param x_suction: Array containing the x-coordinates of the suction side.
        :param y_suction: Array containing the y-coordinates of the suction side.
        :param x_pressure: Array containing the x-coordinates of the pressure side.
        :param y_pressure: Array containing the y-coordinates of the pressure side.
        :return: Arrays containing the coordinates (x and y separately) for each of the sides (suction first,
        then pressure side).
        """
        for _ in np.arange((1 / 5) * self.n_points + 1, (3 / 5) * self.n_points + 1):
            x_pressure = np.append(x_pressure, x_pressure[-1] + self.dxp)
            y_pressure = np.append(y_pressure, self.aP + self.bP * x_pressure[-1] + self.cP * x_pressure[-1] ** 2 +
                                   self.dP * x_pressure[-1] ** 3)

            x_suction = np.append(x_suction, x_suction[-1] + self.dxs)
            y_suction = np.append(y_suction, self.aS + self.bS * x_suction[-1] + self.cS * x_suction[-1] ** 2 +
                                  self.dS * x_suction[-1] ** 3)
        self.dxs = (self.x1 - self.x2) / ((1 / 5) * self.n_points)

        return x_suction, y_suction, x_pressure, y_pressure

    def _define_pressure_suction_3(self, x_suction, y_suction, x_pressure, y_pressure):
        """
        Define the third order polynomial on the pressure side and the circular arc on the suction side. This function
        will separate coords between pressure and suction surfaces.
        :param x_suction: Array containing the x-coordinates of the suction side.
        :param y_suction: Array containing the y-coordinates of the suction side.
        :param x_pressure: Array containing the x-coordinates of the pressure side.
        :param y_pressure: Array containing the y-coordinates of the pressure side.
        :return: Arrays containing the coordinates (x and y separately) for each of the sides (suction first,
        then pressure side).
        """
        for _ in np.arange((3 / 5) * self.n_points + 1, (4 / 5) * self.n_points + 1):
            x_pressure = np.append(x_pressure, x_pressure[-1] + self.dxp)
            y_pressure = np.append(y_pressure, self.aP + self.bP * x_pressure[-1] + self.cP * x_pressure[-1] ** 2 +
                                   self.dP * x_pressure[-1] ** 3)

            x_suction = np.append(x_suction, x_suction[-1] + self.dxs)
            y_suction = np.append(y_suction, self.y0 + np.sqrt(self.R0 ** 2 - (x_suction[-1] - self.x0) ** 2))
        self.dxp = (self.x6 - self.x5) / ((1 / 5) * self.n_points + 1)
        self.dxs = (self.x6 - self.x1) / (1 / 5) * self.n_points + 1

        return x_suction, y_suction, x_pressure, y_pressure

    def _define_pressure_suction_4(self, x_suction, y_suction, x_pressure, y_pressure):
        """
        Define the trailing edge arc joining elementary points 5 and 1. This function will separate coords between
        pressure and suction sides.
        :param x_suction: Array containing the x-coordinates of the suction side.
        :param y_suction: Array containing the y-coordinates of the suction side.
        :param x_pressure: Array containing the x-coordinates of the pressure side.
        :param y_pressure: Array containing the y-coordinates of the pressure side.
        :return: Arrays containing the coordinates (x and y separately) for each of the sides (suction first,
        then pressure side).
        """
        # Define the angle range for the elliptical arc
        theta_range = np.linspace(0, -np.pi, num=int(self.n_points // 2))

        # Calculate the x-coordinates of the trailing edge
        x_te = self.te_r * np.cos(theta_range)

        # Calculate the y-coordinates of the trailing edge
        y_te = self.te_r * np.sin(theta_range)

        # Compute the center coordinates of the half-ellipse
        x_center = (x_suction[-1] + x_pressure[-1]) / 2
        y_center = (y_suction[-1] + y_pressure[-1]) / 2

        # Rotate the points by outlet_blade angle
        angle = vecAngle(np.array([x_suction[-1], x_suction[-1]]),
                         np.array([x_suction[-1] - x_pressure[-1], x_suction[-1] - y_pressure[-1]]))
        # Shift the rotated coordinates back to the original center

        rots = np.sin(angle)
        rotc = np.cos(angle)
        rot = np.array([[rotc, -rots], [rots, rotc]])

        ts_rotated = np.dot(rot, np.stack([x_te, y_te])).T + np.array([x_center, y_center])

        # Shift the rotated coordinates back to the original center
        x_te_rotated = ts_rotated[::, 0]
        y_te_rotated = ts_rotated[::, 1]
        # Append the rotated trailing edge coordinates to the arrays of points
        x_pressure = np.append(x_pressure, x_te_rotated)
        y_pressure = np.append(y_pressure, y_te_rotated)

        x_suction = np.append(x_suction, x_te_rotated)
        y_suction = np.append(y_suction, y_te_rotated)

        return x_suction, y_suction, x_pressure, y_pressure


def create_turbine_profile(radius, axial_chord, tangential_chord, unguided_turning, inlet_blade, inlet_half_wedge,
                           le_r, outlet_blade, te_r, n_blades, throat, straight_te):
    # Create an instance of TurbineBladeGen with the desired parameters
    blade_gen = TurbineBladeGen(radius=radius, axial_chord=axial_chord, tangential_chord=tangential_chord,
                                unguided_turning=unguided_turning, inlet_blade=inlet_blade,
                                inlet_half_wedge=inlet_half_wedge, le_r=le_r, outlet_blade=outlet_blade, te_r=te_r,
                                n_blades=n_blades, throat=throat)

    # Generate blade geometry and obtain coordinates
    blade_gen.calculate_blade_geometry(straight_te=straight_te)

    # Get the generated coordinates
    x_suction, y_suction = blade_gen.x_suction, blade_gen.y_suction
    x_pressure, y_pressure = blade_gen.x_pressure, blade_gen.y_pressure

    # append points to xx_blade and yy_blade
    xx_blade = np.append(x_suction, x_pressure[::-1])
    yy_blade = np.append(y_suction, y_pressure[::-1])
    return xx_blade, yy_blade



import numpy as np


def calcAnisoMatrix(reynoldsstress_tensor):
    rpg = np.linalg.eigh(reynoldsstress_tensor)[1]
    r_ii = sum([rpg[0][0], rpg[1][1], rpg[2][2]])
    anisotropy_matrix = reynoldsstress_tensor / r_ii - np.identity(3) / 3
    return anisotropy_matrix


def calcAnisoEigs(anisotropy_matrix):
    """
    Calculates the eigenvalues and eigenvectors of an anisotropy matrix.

    Args:
        anisotropy_matrix (numpy.ndarray): The anisotropy matrix represented as a 3x3 numpy array.

    Returns:
        tuple: A tuple containing the eigenvalues and eigenvectors.
            - eigenvalues (numpy.ndarray): A 1-dimensional numpy array containing the three eigenvalues
              sorted in descending order.
            - eigenvectors (numpy.ndarray or None): A 2-dimensional numpy array containing the three
              eigenvectors as columns. If all eigenvalues are zero, None is returned.

    Raises:
        ValueError: If the input matrix is not a 3x3 numpy array.

    Notes:
        - The anisotropy matrix should be symmetric.
        - The eigenvalues represent the magnitudes of anisotropy along their corresponding eigenvectors.
        - If all eigenvalues are zero, it indicates an isotropic material.

    Example:
        > anisotropy_matrix = np.array([[3, 1, 2], [1, 4, 5], [2, 5, 6]])
        > eigenvalues, eigenvectors = calcAnisoEigs(anisotropy_matrix)
        > print(eigenvalues)
        array([12.866,  0.174, -2.040])
        > print(eigenvectors)
        array([[ 0.240, -0.856,  0.458],
               [-0.716, -0.513, -0.473],
               [ 0.657, -0.058, -0.752]])
    """

    eigen_val = np.linalg.eig(anisotropy_matrix)
    eigens = list(eigen_val[0])
    eigen_vec = list(eigen_val[1])
    if list(eigens) == [0, 0, 0]:
        return np.zeros(3), None
    max_eigen_idx = eigens.index(max(eigens))
    min_eigen_idx = eigens.index(min(eigens))
    middle = [0, 1, 2]
    middle.remove(max_eigen_idx)
    middle.remove(min_eigen_idx)
    middle = middle[0]

    gamma_1 = eigens[max_eigen_idx]
    gamma_2 = eigens[middle]
    gamma_3 = eigens[min_eigen_idx]

    eigen_val = np.array([gamma_1, gamma_2, gamma_3])
    return eigen_val, eigen_vec


def C_barycentric(R):
    """
    Calculates the barycentric weights for a given anisotropic matrix representing the Reynolds stress tensor.

    Parameters:
    R (numpy.ndarray): Anisotropic matrix representing the Reynolds stress tensor.

    Returns:
    numpy.ndarray: Barycentric weights.

    The function calculates the barycentric weights for a given anisotropic matrix, which represents the
    Reynolds stress tensor. The Reynolds stress tensor characterizes the anisotropy present in turbulent flows
    by describing the correlation between fluctuating velocity components.

    The anisotropic matrix R is a 3x3 symmetric matrix defined as:
        R = | R11  R12  R13 |
            | R12  R22  R23 |
            | R13  R23  R33 |

    The function performs the following steps:
    1. Checks if all eigenvalues of the anisotropic matrix are zero. If yes, returns [0, 0, 1] as barycentric weights.
    2. Calculates the eigenvalues of the anisotropic matrix, anisoEigs, which correspond to the magnitudes of the
    principal axes of the Reynolds stress tensor.
    3. Extracts the eigenvalues gamma_1, gamma_2, and gamma_3 from anisoEigs.
    4. Computes the barycentric weights C1c, C2c, and C3c using the eigenvalues.
    5. Constructs and returns the barycentric weights as a numpy array, CWeights.

    Note:
    - The anisotropic matrix R should be a 3x3 numpy array representing the Reynolds stress tensor.
    - The function assumes the existence of the calcAnisoEigs function to calculate the eigenvalues.
    - Make sure to import the necessary modules, such as numpy, for the function to work properly.
    """

    aniso = calcAnisoMatrix(R)
    anisoEigs = calcAnisoEigs(aniso)[0]
    if list(anisoEigs) == [0, 0, 0]:
        return np.array([0, 0, 1])
    else:
        gamma_1 = anisoEigs[0]
        gamma_2 = anisoEigs[1]
        gamma_3 = anisoEigs[2]

    C1c = gamma_1 - gamma_2
    C2c = 2 * (gamma_2 - gamma_3)
    C3c = 3 * gamma_3 + 1
    CWeights = np.array([C1c, C2c, C3c])

    return CWeights


def autocorr(signal):
    """
    Calculate the normalized autocorrelation of a one-dimensional signal.

    Parameters:
    signal (numpy.ndarray): Input signal.

    Returns:
    numpy.ndarray: Normalized autocorrelation.

    Compute the normalized autocorrelation of the input signal by cross-correlating it with itself.
    The result is normalized by dividing by the squared sum of the signal.
    """
    norm = np.sum(np.array(signal) ** 2)
    result = np.correlate(np.array(signal), np.array(signal), 'full') / norm
    return result[int(len(result) / 2):]


def zero_crossings(data_series):
    zcs = np.where(np.diff(np.sign(data_series)))[0]
    return zcs


def reldiff(a, b):
    """Calculates the relative difference between two values or arrays of values.

    Parameters:
    a (float or numpy array): The first value or array of values to compare.
    b (float or numpy array): The second value or array of values to compare.

    Returns:
    float or numpy array: The relative difference between a and b. If a and b are both numpy arrays,
    the output will be a numpy array of the same shape. If one input is a float and the other is a numpy array,
    the output will be a numpy array of the same shape as the input array.

    Notes:
    The relative difference is defined as the absolute difference between the two values divided by the average
    of their absolute values. If a and b are equal, the relative difference is 0. If either a or b is zero,
    the absolute difference between the two values is returned instead of the relative difference.
    """
    if isinstance(a, np.ndarray) or isinstance(b, np.ndarray):
        # Convert to numpy arrays if necessary
        if isinstance(a, float):
            a = np.full(b.shape, a)
        if isinstance(b, float):
            b = np.full(a.shape, b)
        # Calculate relative difference element-wise
        diff = np.abs(a - b) / ((np.abs(a) + np.abs(b)) / 2)
        # Set elements where a or b is zero to the absolute difference
        mask = np.logical_or(a == 0, b == 0)
        diff[mask] = np.abs(a[mask] - b[mask])
        return diff
    else:
        # Handle individual float values
        if a == b:
            return 0
        elif a == 0 or b == 0:
            return abs(a - b)
        else:
            return abs(a - b) / ((abs(a) + abs(b)) / 2)


def return_intersection(hist_1, hist_2):
    """
    Calculate the intersection of two histograms.

    Parameters
    ----------
    hist_1: numpy.ndarray
        The first histogram to be compared.
    hist_2: numpy.ndarray
        The second histogram to be compared.

    Returns
    -------
    float
        The intersection of the two histograms.
    """
    minima = np.minimum(hist_1, hist_2)
    intersection = np.true_divide(np.sum(minima), np.sum(hist_2))
    return intersection


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def is_equidistant(arr, tolerance=1e-8):
    # check if the array has fewer than 2 elements
    if len(arr) < 2:
        return False

    # calculate the difference between the first two elements
    diff = arr[1] - arr[0]

    # create an array of the differences between each element and the next element
    diffs = np.abs(np.diff(arr))

    # use numpy.isclose to check if all of the differences are equal within the tolerance
    return np.all(np.isclose(diffs, diff, atol=tolerance))


def autocorrelate(series):
    """Compute the autocorrelation function of a given time series.

    Args:
        series (array-like): The 1D time series to be autocorrelated.

    Returns:
        array-like: The autocorrelation function of the input series.
    """
    series = series - np.mean(series)
    norm = np.sum(np.array(series) ** 2)
    result = np.correlate(np.array(series), np.array(series), 'full') / norm
    accr = result[int(len(result) / 2):]
    return accr


def minmax_normalize(series_array):
    """Normalize a given 1D array to the range [0, 1] using min-max normalization.

    Args:
        series_array (array-like): The 1D array to be normalized.

    Returns:
        array-like: The normalized 1D array with values in the range [0, 1].

    Note:
        If all values in the input array are the same, the function will return the
        input array unchanged.
    """

    if not np.all(series_array == series_array[0]):
        mins = min(series_array)
        maxs = max(series_array)
        f = (-mins + series_array) / (maxs - mins)
        return f
    else:
        return series_array



def test_aspect_ratio():
    from ntrfc.meshquality.aspect_ratio import compute_cell_aspect_ratios
    import pyvista as pv
    import numpy as np
    grid = pv.UniformGrid()
    values = np.linspace(0, 10, 1000).reshape((20, 5, 10))
    values.shape
    # Set the grid dimensions: shape + 1 because we want to inject our values on
    #   the CELL data
    grid.dimensions = np.array(values.shape) + 1

    # Edit the spatial reference
    grid.origin = (100, 33, 55.6)  # The bottom left corner of the data set
    grid.spacing = (1, 5, 2)  # These are the cell sizes along each axis

    aspect_ratios = compute_cell_aspect_ratios(grid)

    assert np.all(aspect_ratios == max(grid.spacing) / min(grid.spacing))



def test_calcmidpassagestreamline():
    from ntrfc.turbo.pointcloud_methods import extract_geo_paras
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.turbo.cascade_geometry import calcmidpassagestreamline

    import numpy as np
    import pyvista as pv

    naca_code = "6009"
    angle = 10  # deg
    alpha = 1
    res = 240
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    sorted_poly.rotate_z(angle, inplace=True)

    poly, ps_poly, ss_poly, ind_hk, ind_vk, mids_poly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
        sorted_poly, alpha)

    calcmidpassagestreamline(mids_poly.points[::, 0], mids_poly.points[::, 1], beta_leading, beta_trailing, -1, 2, 1)



import numpy as np
import pyvista as pv


def test_polyline_from_points():
    from ntrfc.geometry.line import polyline_from_points

    points = pv.Line(resolution=100).points
    line = polyline_from_points(points)
    assert line.length == 1.0, "theres something fishy about the polyline_from_points implementation"


def test_line_from_points():
    from ntrfc.geometry.line import lines_from_points

    points = pv.Line(resolution=100).points
    line = lines_from_points(points)
    assert line.length == 1.0, "theres something fishy about the polyline_from_points implementation"


def test_refine_spline():
    """
    tests if you can refine a spline by checking the number of points and the length of the spline
    """
    from ntrfc.geometry.line import refine_spline

    coarseres = 2
    line = pv.Line(resolution=coarseres)
    fineres = 100
    fline_xx, fline_yy = refine_spline(line.points[::, 0], line.points[::, 1], fineres)
    fline = pv.lines_from_points(np.stack([fline_xx, fline_yy, np.zeros(len(fline_xx))]).T)
    assert line.length == fline.length
    assert fline.number_of_points == fineres

    coarseres = 2
    line = pv.Line(resolution=coarseres)
    fineres = 100
    fline_xx, fline_yy = refine_spline(line.points[::, 0], line.points[::, 1], fineres, half_cosine_spacing=True)
    fline = pv.lines_from_points(np.stack([fline_xx, fline_yy, np.zeros(len(fline_xx))]).T)
    assert line.length == fline.length
    assert fline.number_of_points == fineres



import os

from ntrfc.filehandling.datafiles import inplace_change, get_filelist_fromdir, get_directory_structure


def test_yamlDictRead(tmpdir):
    """
    tests if yaml is returning a known dictionary
    """
    from ntrfc.filehandling.datafiles import yaml_dict_read

    test_file = tmpdir / "test.yaml"
    with open(test_file, "w") as handle:
        handle.write("test_key: True\n")
    assert yaml_dict_read(test_file) == {"test_key": True}


def test_yamlDictWrite(tmpdir):
    """
    tests if yaml is writing and returning a known dictionary
    """
    from ntrfc.filehandling.datafiles import yaml_dict_read, write_yaml_dict

    test_file = tmpdir / "test.yaml"
    test_dict = {"test_key": True}
    write_yaml_dict(test_file, test_dict)

    assert yaml_dict_read(test_file) == test_dict


def test_pickle_operations(tmpdir):
    """
    checks if the pickle-operators are working
    :param tmpdir:
    :return:
    """
    from ntrfc.filehandling.datafiles import write_pickle, read_pickle, write_pickle_protocolzero

    fname = tmpdir / "test.pkl"
    dict = {"test": 1}
    write_pickle(fname, dict)
    pklread = read_pickle(fname)
    assert dict["test"] == pklread["test"]
    write_pickle_protocolzero(fname, dict)
    pklread_zero = read_pickle(fname)
    assert dict["test"] == pklread_zero["test"]


def test_create_dirstructure(tmpdir):
    from ntrfc.filehandling.datafiles import create_dirstructure

    dirstructure = ["ast/bla", "ast/ble"]
    create_dirstructure(dirstructure, tmpdir)
    checks = [os.path.isdir(path) for path in [os.path.join(tmpdir, relpath) for relpath in dirstructure]]
    assert all(checks), "not all directories have been created"


def test_inplace_change(tmpdir):
    test_oldstring = "old"
    test_newstring = "new"
    test_file = tmpdir / "file.txt"
    with open(test_file, "w") as fobj:
        fobj.write(test_oldstring)

    inplace_change(test_file, test_oldstring, test_newstring)

    with open(test_file, "r") as fobj:
        ans = fobj.read()

    assert ans == test_newstring, "inplace change not working"


def test_get_filelist_fromdir(tmpdir):
    os.mkdir(f"{tmpdir}/somedir")

    file_one = "somedir/somefile"
    file_two = "somefile"
    open(f"{tmpdir}/{file_one}", "w")
    open(f"{tmpdir}/{file_two}", "w")
    filelist = get_filelist_fromdir(tmpdir)
    assert len(filelist) == 2
    filebase = [os.path.relpath(i, tmpdir) for i in filelist]
    assert file_one in filebase
    assert file_two in filebase


def test_get_directory_structure(tmpdir):
    # create a directory structure in a temporary directory
    tmpdir.mkdir('root')
    tmpdir.join('root/folder1').mkdir()
    tmpdir.join('root/folder2').mkdir()
    tmpdir.join('root/folder1/subfolder1').mkdir()
    tmpdir.join('root/folder1/subfolder2').mkdir()
    tmpdir.join('root/folder1/subfolder2/file1.txt').write('test')
    tmpdir.join('root/folder1/subfolder2/file2.txt').write('test')
    tmpdir.join('root/folder1/file3.txt').write('test')
    tmpdir.join('root/folder2/file4.txt').write('test')

    # get the directory structure
    directory_structure = get_directory_structure(str(tmpdir.join('root')))

    # check that the directory structure is correct
    assert directory_structure == {
        'root': {'folder1': {'file3.txt': None,
                             'subfolder1': {},
                             'subfolder2': {'file1.txt': None, 'file2.txt': None}},
                 'folder2': {'file4.txt': None}}}



import os
import tempfile
from dataclasses import dataclass

import numpy as np
import pyvista as pv

from ntrfc.cascade_case.casemeta.casemeta import CaseMeta
from ntrfc.cascade_case.utils.domain_utils import DomainParameters
from ntrfc.math.vectorcalc import vecAbs
from ntrfc.turbo.pointcloud_methods import calcMidPassageStreamLine


@dataclass
class CascadeDomain2D:
    casemeta: CaseMeta = CaseMeta(tempfile.mkdtemp())
    pressureside: pv.PolyData = None
    suctionside: pv.PolyData = None
    profilepoints: pv.PolyData = None
    le_index: int = None
    te_index: int = None
    beta_leading: float = None
    beta_trailing: float = None
    yperiodic_low: pv.PolyData = None
    yperiodic_high: pv.PolyData = None
    inlet: pv.PolyData = None
    outlet: pv.PolyData = None
    pitch: float = None
    chordlength: float = None

    def generate_from_cascade_parameters(self, domainparams: DomainParameters):
        # Use params attributes to generate attributes of CascadeDomain2D
        self.profilepoints = domainparams.profile_points
        self.le_index = domainparams.leading_edge_index
        self.te_index = domainparams.trailing_edge_index
        x_mids = domainparams.midspoly.points[::, 0]
        y_mids = domainparams.midspoly.points[::, 1]
        self.beta_leading = domainparams.beta_in
        self.beta_trailing = domainparams.beta_out
        x_inlet = domainparams.xinlet
        x_outlet = domainparams.xoutlet
        self.pitch = domainparams.pitch
        blade_shift = domainparams.blade_yshift

        x_mpsl, y_mpsl = calcMidPassageStreamLine(x_mids, y_mids, self.beta_leading, self.beta_trailing,
                                                  x_inlet, x_outlet, self.pitch)

        y_upper = np.array(y_mpsl) + blade_shift
        per_y_upper = pv.lines_from_points(np.stack((np.array(x_mpsl),
                                                     np.array(y_upper),
                                                     np.zeros(len(x_mpsl)))).T)
        y_lower = y_upper - self.pitch
        per_y_lower = pv.lines_from_points(np.stack((np.array(x_mpsl),
                                                     np.array(y_lower),
                                                     np.zeros(len(x_mpsl)))).T)

        inlet_pts = np.array([per_y_lower.points[per_y_lower.points[::, 0].argmin()],
                              per_y_upper.points[per_y_upper.points[::, 0].argmin()]])

        inletPoly = pv.Line(*inlet_pts)
        outlet_pts = np.array([per_y_lower.points[per_y_lower.points[::, 0].argmax()],
                               per_y_upper.points[per_y_upper.points[::, 0].argmax()]])

        outletPoly = pv.Line(*outlet_pts)

        self.pressureside = domainparams.pspoly
        self.suctionside = domainparams.sspoly
        self.yperiodic_low = per_y_lower
        self.yperiodic_high = per_y_upper
        self.inlet = inletPoly
        self.outlet = outletPoly
        self.chordlength = vecAbs(self.profilepoints.points[self.te_index] - self.profilepoints.points[self.le_index])

    def plot_domain(self):
        """
        Plot the domain parameters using PyVista.


        Returns:
            pv.Plotter: The PyVista plotter object used for plotting.
        """
        if os.getenv('DISPLAY') is None:
            pv.start_xvfb()  # Start X virtual framebuffer (Xvfb)
        plotter = pv.Plotter(off_screen=True)
        path = os.path.join(self.casemeta.case_root_directory, "domain.png")

        plotter.window_size = 2400, 2400
        # Plot the suction side and pressure side polys
        plotter.add_mesh(self.suctionside, color='b', show_edges=True)
        plotter.add_mesh(self.pressureside, color='r', show_edges=True)
        plotter.add_mesh(self.yperiodic_low)
        plotter.add_mesh(self.yperiodic_high)
        plotter.add_mesh(self.inlet)
        plotter.add_mesh(self.outlet)

        plotter.add_axes()
        plotter.view_xy()
        plotter.screenshot(path)
        return path



import os

ON_CI = 'CI' in os.environ


def test_cascade_2d_domain():
    import pyvista as pv
    import numpy as np
    from ntrfc.cascade_case.utils.domain_utils import DomainParameters
    from ntrfc.cascade_case.domain import CascadeDomain2D
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    xs, ys = naca("6510", 256)
    points = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    alpha = 1
    domainparas = DomainParameters()
    domainparas.generate_params_by_pointcloud(points, alpha=alpha)
    domainparas.xinlet = -3
    domainparas.xoutlet = 4
    domainparas.pitch = 2
    domainparas.blade_yshift = 0.1
    domainparas.plot_domainparas()
    domain2d = CascadeDomain2D()
    domain2d.generate_from_cascade_parameters(domainparas)
    domain2d.plot_domain()



import numpy as np
from scipy.optimize import least_squares


def func_p(x0, a, b, c, A):
    x = x0
    F = A * ((a ** x) - 1) / ((b ** x) - 1) * (b ** x) - c  # least squares von F ist p
    return F


def calcp(x0, A, N1, N2, N3, fc1, fc2, fc3, D):
    a = (N2 / N3) ** (1 / D)
    b = (N1 / N2) ** (1 / D)
    c = (fc3 - fc2) / (fc2 - fc1)

    p = least_squares(func_p, x0, args=(a, b, c, A))

    return p.x[0]


def prel(p_opt, P):
    p_rel = abs((p_opt - P) / p_opt)
    return p_rel


def getGCI(N1, N2, N3, fc1, fc2, fc3, D, Fs=1.25):
    assert not (((fc2 > fc3) & (fc2 > fc1)) | ((fc2 < fc3) & (fc2 < fc1))), "values are not monotone"
    assert not ((N1 <= N2) or (N2 <= N3)), "Number of cells must be N1 > N2 > N3"

    r21 = (N1 / N2) ** (1 / D)
    r32 = (N2 / N3) ** (1 / D)

    epsilon32 = (fc3 - fc2)
    epsilon21 = (fc2 - fc1)

    x0 = np.log(epsilon32 / epsilon21) / np.log(r21)
    p = calcp(x0, 1, N1, N2, N3, fc1, fc2, fc3, D)

    GCI_1 = Fs * abs((epsilon21 / fc1)) * 1 / ((r21 ** p) - 1)
    GCI_2 = Fs * abs((epsilon32 / fc2)) * 1 / ((r32 ** p) - 1)
    GCI_3 = GCI_2 * r32 ** p

    f_extra = fc1 + (fc1 - fc2) / ((r21 ** p) - 1)

    EERE_1 = abs((f_extra - fc1) / f_extra)
    EERE_2 = abs((f_extra - fc2) / f_extra)
    EERE_3 = abs((f_extra - fc3) / f_extra)

    A_Flag = GCI_2 / (GCI_1 * (r21 ** p))

    if ((A_Flag > 1.15) | (A_Flag < 0.85)):
        print('Error: Your fc-values are not in the asymptotic range!')
        print('Refine your grid and repeat grid convergence study, or contact your CFD-expert')

    GCI_1_p1 = Fs * abs((epsilon21 / fc1)) * 1 / ((r21) - 1)
    GCI_2_p1 = Fs * abs((epsilon32 / fc2)) * 1 / ((r32) - 1)
    GCI_3_p1 = GCI_2_p1 * r32

    f_extra_p1 = fc1 + (fc1 - fc2) / ((r21) - 1)

    EERE_1_p1 = abs((f_extra_p1 - fc1) / f_extra_p1)
    EERE_2_p1 = abs((f_extra_p1 - fc2) / f_extra_p1)
    EERE_3_p1 = abs((f_extra_p1 - fc3) / f_extra_p1)

    text = ""
    text += f'{"GCI_1": <16} {GCI_1}\n'
    text += f'{"GCI_2": <16} {GCI_2}\n'
    text += f'{"GCI_3": <16} {GCI_3}\n'
    text += f'{"EERE_1": <16} {EERE_1}\n'
    text += f'{"EERE_2": <16} {EERE_2}\n'
    text += f'{"EERE_3": <16} {EERE_3}\n'
    text += f'{"GCI_1_p1": <16} {GCI_1_p1}\n'
    text += f'{"GCI_2_p1": <16} {GCI_2_p1}\n'
    text += f'{"GCI_3_p1": <16} {GCI_3_p1}\n'
    text += f'{"EERE_1_p1": <16} {EERE_1_p1}\n'
    text += f'{"EERE_2_p1": <16} {EERE_2_p1}\n'
    text += f'{"EERE_3_p1": <16} {EERE_3_p1}\n'
    print(text)
    return GCI_1, GCI_2, GCI_3
    # return [GCI_1, GCI_2, GCI_3], [EERE_1, EERE_2, EERE_3], [GCI_1_p1, GCI_2_p1, GCI_3_p1], \
    #        [EERE_1_p1, EERE_2_p1, EERE_3_p1]



import numpy as np


def test_total_pressure():
    from ntrfc.fluid.fluid import total_pressure
    # Test input where kappa = 1.4, mach_number = 0.5, pressure = 100
    expected_output = 189292.91587378542
    output = total_pressure(1.4, 1, 100000)
    assert np.isclose(expected_output, output, rtol=1e-10, atol=1e-10)


def test_mach_number():
    from ntrfc.fluid.fluid import mach_number
    # Test input where c = 340, kappa = 1.4, R_L = 287, T = 273
    expected_output = 1.02
    output = mach_number(340.3, 1.4, 287, 273)
    assert abs(output - expected_output) < 1e-2


def test_sutherland_viscosity():
    from ntrfc.fluid.fluid import sutherland_viscosity
    temperature = 300  # degrees Kelvin
    expected_dynamic_viscosity = 1.846e-05  # square meters per second

    # Call the Sutherland_Law function
    dynamic_viscosity = sutherland_viscosity(temperature)

    # Check if the output is close to the expected output
    assert np.isclose(dynamic_viscosity, expected_dynamic_viscosity, rtol=1e-10, atol=1e-10)


def test_ma_is():
    from ntrfc.fluid.isentropic import ma_is
    # Test input where outflow_static_pressure = 100, isentropic_exponent = 1.4, pressure = 50, velocity = 20,
    # gas_constant = 287, temperature = 300
    expected_output = 1.71
    output = ma_is(10000, 1.4, 50000, 20, 287, 300)
    assert abs(output - expected_output) < 1e-2


def test_isentropic_reynolds_number():
    from ntrfc.fluid.isentropic import isentropic_reynolds_number
    # Test input where kappa = 1.4, specific_gas_constant = 287, chord_length = 1, sutherland_reference_viscosity = 1.46e-5,
    # mach_number = 0.65, pressure = 50, temperature = 300, sutherland_reference_temperature = 110.4
    expected_output = 708.95
    output = isentropic_reynolds_number(1.4, 287, 1, 1.46e-5, 0.65, 50, 300, 110.4)
    assert abs(output - expected_output) < 1e-2


def test_isentropic_total_temperature():
    from ntrfc.fluid.isentropic import isentropic_total_temperature
    # Test input where kappa = 1.4, mach_number = 0.5, temperature = 300
    expected_output = 315
    output = isentropic_total_temperature(1.4, 0.5, 300)
    assert abs(output - expected_output) < 1e-2



def test_parsec():
    from ntrfc.turbo.airfoil_generators.parsec_airfoil_creator import parsec_airfoil_gen

    R_LE = 0.01
    x_PRE = 0.450
    y_PRE = -0.006
    d2y_dx2_PRE = -0.4
    th_PRE = 0.05
    x_SUC = 0.450
    y_SUC = 0.055
    d2y_dx2_SUC = -0.350
    th_SUC = -6

    pparray = [R_LE, x_PRE, y_PRE, d2y_dx2_PRE, th_PRE, x_SUC, y_SUC, d2y_dx2_SUC, th_SUC]
    profile_points = parsec_airfoil_gen(pparray)
    x, y = profile_points[::, 0], profile_points[::, 1]
    assert len(x) == len(y)



import numpy as np


def avdr(rho_1, mag_u_1, beta_1, rho_2, mag_u_2, beta_2):
    """
    :param rho_1: float
    :param mag_u_1: float
    :param beta_1: float
    :param rho_2: float
    :param mag_u_2: float
    :param beta_2: float
    :return: float
    """
    if beta_2 == beta_1:
        avdr_res = rho_2 * mag_u_2 / (rho_1 * mag_u_1)
    else:
        avdr_res = rho_2 * mag_u_2 * np.sin(beta_2) / (rho_1 * mag_u_1 * np.sin(beta_1))
    return avdr_res



import numpy as np

from ntrfc.turbo.airfoil_generators.parametric_turbine_generator import create_turbine_profile


def test_create_turbine_profile():
    radius = 5.5
    axial_chord = 1.102
    tangential_chord = 0.591
    unguided_turning = 6.5 * np.pi / 180
    inlet_blade = 35 * np.pi / 180
    inlet_half_wedge = 9 * np.pi / 180
    le_r = 0.0310
    outlet_blade = -57 * np.pi / 180
    te_r = 0.0160
    n_blades = 51
    throat = 0.337
    straight_te = False

    xs, ys = create_turbine_profile(radius, axial_chord, tangential_chord,
                                    unguided_turning, inlet_blade,
                                    inlet_half_wedge,
                                    le_r, outlet_blade, te_r, n_blades, throat,
                                    straight_te)

    assert len(xs) == len(ys)



import numpy as np
from matplotlib import path as mpltPath

from ntrfc.math.vectorcalc import vecAbs, vecProjection


def massflow_plane(mesh, rhoname="rho", velocityname="U"):
    if not "Normals" in mesh.array_names:
        mesh = mesh.compute_normals()
    if not "Area" in mesh.array_names:
        mesh = mesh.compute_cell_sizes()
    mesh = mesh.point_data_to_cell_data()
    normals = mesh.cell_normals
    rhos = mesh[rhoname]
    areas = mesh["Area"]
    velocities = mesh[velocityname]

    massflow = np.array(
        [vecAbs(vecProjection(velocities[i], normals[i])) for i in range(mesh.number_of_cells)]) ** 2 * rhos * areas

    return massflow


def massflowave_plane(mesh, valname, rhoname="rho", velocityname="U"):
    massflow = massflow_plane(mesh, rhoname=rhoname, velocityname=velocityname)

    mass_ave = np.sum(mesh[valname] * massflow) / np.sum(massflow)

    return mass_ave


def areaave_plane(mesh, valname):
    array = mesh[valname]
    if not "Area" in mesh.array_names:
        mesh = mesh.compute_cell_sizes()
    areas = mesh["Area"]
    area_ave = np.sum((array.T * areas).T) / np.sum(areas)
    return area_ave


def inside_poly(polygon, points):
    """
    :param polygon: (x,y)
    :param points: (x,y)
    :return: list of True or False for indicating weather inside or not
    """
    path = mpltPath.Path(polygon)
    inside = path.contains_points(points)
    return inside



import os
import pickle
from functools import reduce
from pathlib import Path

import yaml


def yaml_dict_read(yml_file):
    args_from_yaml = {}

    with open(yml_file, "r", newline='') as Fobj:
        document = yaml.load_all(Fobj, Loader=yaml.FullLoader)
        for settings in document:
            for key, value in settings.items():
                args_from_yaml[key] = value
    return args_from_yaml


def write_pickle_protocolzero(file, args):
    with open(file, "wb") as Fobj:
        pickle.dump(args, Fobj, protocol=0)


def write_pickle(file, args):
    with open(file, "wb") as Fobj:
        pickle.dump(args, Fobj)


def read_pickle(file):
    with open(file, 'rb') as f:
        args = pickle.load(f)
    return args


def write_yaml_dict(fpath, data):
    """
    :param fpath: target path
    :param data: dictionary
    """
    with open(fpath, 'w') as outfile:
        yaml.dump(data, outfile, default_flow_style=False)


def inplace_change(filename, old_string, new_string):
    # Safely read the input filename using 'with'
    with open(filename) as f:
        s = f.read()
        if old_string not in s:
            # print('"{old_string}" not found in {filename}.'.format(**locals()))
            return

    # Safely write the changed content, if found in the file
    with open(filename, 'w') as f:
        print('Inserting "{old_string}" to "{new_string}" in {filename}'.format(**locals()))
        s = s.replace(old_string, new_string)
        f.write(s)


def get_directory_structure(rootdir):
    """
    Creates a nested dictionary representing the folder structure of the specified root directory.

    Args:
        rootdir (str): The root directory to traverse.

    Returns:
        dict: A nested dictionary representing the folder structure of the root directory. The keys
              are the folder names and the values are dictionaries for subdirectories or None for files.

    Example:
        get_directory_structure('/home/user/documents')
        {
            'documents': {
                'folder1': {
                    'subfolder1': None,
                    'subfolder2': {
                        'file1.txt': None,
                        'file2.txt': None
                    },
                    'file3.txt': None
                },
                'folder2': {
                    'file4.txt': None
                }
            }
        }
    """
    dir = {}
    rootdir = os.path.join(rootdir)
    rootdir = rootdir.rstrip(os.sep)
    start = rootdir.rfind(os.sep) + 1
    for path, dirs, files in os.walk(rootdir):
        folders = path[start:].split(os.sep)
        subdir = dict.fromkeys(files)
        parent = reduce(dict.get, folders[:-1], dir)
        parent[folders[-1]] = subdir
    return dir


def get_filelist_fromdir(path):
    filelist = []
    for r, d, f in os.walk(path):
        for file in f:
            filelist.append(os.path.join(r, file))
    return filelist


def create_dirstructure(directories, path):
    """

    :param : directories - list of directories
    :param : path - path
    """
    for d in directories:
        Path(os.path.join(path, d)).mkdir(parents=True, exist_ok=True)
    return 0



import os
from dataclasses import dataclass, field

import pyvista as pv
from tqdm import tqdm

from ntrfc.filehandling.mesh import load_mesh


@dataclass
class SliceSeries:
    """
    object for storing pathes to slices
    """
    slicesets: list = field(default_factory=list)

    def add_sliceset(self, slices, groupname, timestamps):
        self.slicesets.append(sliceset(slices, groupname, timestamps))

    def create_animation(self, slicegroup, variable, path, name):
        slicegroups = [self.slicesets[i].groupname for i in range(len(self.slicesets))]
        use_idx = slicegroups.index(slicegroup)
        use_slices = self.slicesets[use_idx].slices
        use_timestamps = self.slicesets[use_idx].timestamps

        pv.set_plot_theme("document")
        fontsize = 36
        labelsize = 36

        gifpath = os.path.join(path, name)

        first_slice = load_mesh(use_slices[0])

        p = pv.Plotter(notebook=False, off_screen=True)
        p.window_size = [3840, 2160]
        p.open_gif(gifpath)
        p.add_mesh(first_slice, scalars=variable)
        p.add_text(f"framenumber: 0 - time {use_timestamps[0]}", font_size=labelsize, position='upper_left')
        p.camera.position = (0, 0, 1)
        p.camera.roll = 0
        p.camera.zoom = 1

        for idx, (path, time) in tqdm(enumerate(zip(use_slices[1:], use_timestamps[1:]))):
            sliceframe = load_mesh(path)
            p.clear()
            p.add_mesh(sliceframe, scalars=variable)
            p.add_text(f"framenumber: {idx} - time {time}", font_size=fontsize, position='upper_left')
            p.scalar_bar.SetTitleRatio(2)
            p.write_frame()

        # Closes and finalizes movie
        p.close()


@dataclass
class sliceset:
    '''Object for tracking probes a set.'''
    slices: list
    groupname: str
    timestamps: list
